#!/usr/bin/python
"""functions that interact with the C code, for handling convolutions mostly.
"""
import ctypes
import warnings
import os
import glob
import numpy as np

# the wrapConv.so file can have some system information after it from the compiler, so we just find
# whatever it is called
lib_folder = os.path.dirname(os.path.realpath(__file__))
libpath = glob.glob(os.path.join(lib_folder, 'wrapConv*.so')) + \
    glob.glob(os.path.join(lib_folder, 'wrapConv.*.pyd'))
# print(libpath)

# load the c library
if len(libpath) > 0:
    lib = ctypes.cdll.LoadLibrary(libpath[0])
else:
    warnings.warn("Can't load in C code, something went wrong in your install!")


def corrDn(image, filt, edge_type='reflect1', step=(1, 1), start=(0, 0), stop=None):
    """Compute correlation of image with filt, followed by downsampling.

    These arguments should be 1D or 2D arrays, and image must be larger (in both dimensions) than
    filt.  The origin of filt is assumed to be floor(size(filt)/2)+1.

    Downsampling factors are determined by step (optional, default=(1, 1)), which should be a
    2-tuple (y, x).

    The window over which the convolution occurs is specfied by start (optional, default=(0,0), and
    stop (optional, default=size(image)).

    NOTE: this operation corresponds to multiplication of a signal vector by a matrix whose rows
    contain copies of the filt shifted by multiples of step.  See `upConv` for the operation
    corresponding to the transpose of this matrix.

    WARNING: if both the image and filter are 1d, they must be 1d in the same dimension. E.g., if
    image.shape is (1, 36), then filt.shape must be (1, 5) and NOT (5, 1). If they're both 1d and
    1d in different dimensions, then this may encounter a segfault. I've not been able to find a
    way to avoid that within this function (simply reshaping it does not work).

    Arguments
    ---------
    image : `array_like`
        1d or 2d array containing the image to correlate and downsample.
    filt : `array_like`
        1d or 2d array containing the filter to use for correlation and downsampling.
    edge_type : {'circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend', 'dont-compute'}
        Specifies how to handle edges. Options are:

        * `'circular'` - circular convolution
        * `'reflect1'` - reflect about the edge pixels
        * `'reflect2'` - reflect, doubling the edge pixels
        * `'repeat'` - repeat the edge pixels
        * `'zero'` - assume values of zero outside image boundary
        * `'extend'` - reflect and invert
        * `'dont-compute'` - zero output when filter overhangs imput boundaries.
    step : `tuple`
        2-tuple (y, x) which determines the downsampling factor
    start : `tuple`
        2-tuple which specifies the start of the window over which we perform the convolution
    start : `tuple` or None
        2-tuple which specifies the end of the window over which we perform the convolution. If
        None, perform convolution over the whole image

    Returns
    -------
    result : `np.array`
        the correlated and downsampled array

    """
    image = image.copy().astype(float)
    filt = filt.copy().astype(float)

    if image.shape[0] < filt.shape[0] or image.shape[1] < filt.shape[1]:
        raise Exception("Signal smaller than filter in corresponding dimension: ", image.shape, filt.shape, " see parse filter")

    if edge_type not in ['circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend', 'dont-compute']:
        raise Exception("Don't know how to do convolution with edge_type %s!" % edge_type)

    if filt.ndim == 1:
        filt = filt.reshape(1, -1)

    if stop is None:
        stop = (image.shape[0], image.shape[1])

    rxsz = len(range(start[0], stop[0], step[0]))
    rysz = len(range(start[1], stop[1], step[1]))
    result = np.zeros((rxsz, rysz))

    if edge_type == 'circular':
        lib.internal_wrap_reduce(image.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                                 image.shape[1], image.shape[0],
                                 filt.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                                 filt.shape[1], filt.shape[0],
                                 start[1], step[1], stop[1], start[0], step[0],
                                 stop[0],
                                 result.ctypes.data_as(ctypes.POINTER(ctypes.c_double)))
    else:
        tmp = np.zeros((filt.shape[0], filt.shape[1]))
        lib.internal_reduce(image.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                            image.shape[1], image.shape[0],
                            filt.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                            tmp.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                            filt.shape[1], filt.shape[0],
                            start[1], step[1], stop[1], start[0], step[0],
                            stop[0],
                            result.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                            edge_type.encode('ascii'))

    return result


def upConv(image, filt, edge_type='reflect1', step=(1, 1), start=(0, 0), stop=None):
    """Upsample matrix image, followed by convolution with matrix filt.

    These arguments should be 1D or 2D matrices, and image must be larger (in both dimensions) than
    filt.  The origin of filt is assumed to be floor(size(filt)/2)+1.

    Upsampling factors are determined by step (optional, default=(1, 1)),
    a 2-tuple (y, x).

    The window over which the convolution occurs is specfied by start (optional, default=(0, 0),
    and stop (optional, default = step .* (size(IM) + floor((start-1)./step))).

    NOTE: this operation corresponds to multiplication of a signal vector by a matrix whose columns
    contain copies of the time-reversed (or space-reversed) FILT shifted by multiples of STEP.  See
    corrDn.m for the operation corresponding to the transpose of this matrix.

    WARNING: if both the image and filter are 1d, they must be 1d in the same dimension. E.g., if
    image.shape is (1, 36), then filt.shape must be (1, 5) and NOT (5, 1). If they're both 1d and
    1d in different dimensions, then this may encounter a segfault. I've not been able to find a
    way to avoid that within this function (simply reshaping it does not work).

    Arguments
    ---------
    image : `array_like`
        1d or 2d array containing the image to upsample and convolve.
    filt : `array_like`
        1d or 2d array containing the filter to use for upsampling and convolution.
    edge_type : {'circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend', 'dont-compute'}
        Specifies how to handle edges. Options are:

        * `'circular'` - circular convolution
        * `'reflect1'` - reflect about the edge pixels
        * `'reflect2'` - reflect, doubling the edge pixels
        * `'repeat'` - repeat the edge pixels
        * `'zero'` - assume values of zero outside image boundary
        * `'extend'` - reflect and invert
        * `'dont-compute'` - zero output when filter overhangs imput boundaries.
    step : `tuple`
        2-tuple (y, x) which determines the upsampling factor
    start : `tuple`
        2-tuple which specifies the start of the window over which we perform the convolution.
    start : `tuple` or None
        2-tuple which specifies the end of the window over which we perform the convolution. If
        None, perform convolution over the whole image

    Returns
    -------
    result : `np.array`
        the correlated and downsampled array

    """
    image = image.copy().astype(float)
    filt = filt.copy().astype(float)

    if image.ndim == 1:
        image = image.reshape(-1, 1)

    image_shape = (image.shape[0] * step[0], image.shape[1] * step[1])

    if image_shape[0] < filt.shape[0] or image_shape[1] < filt.shape[1]:
        raise Exception("Signal smaller than filter in corresponding dimension: ", image_shape, filt.shape, " see parse filter")

    if edge_type not in ['circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend',
                         'dont-compute']:
        raise Exception("Don't know how to do convolution with edge_type %s!" % edge_type)

    # from upConv.c, the c code that gets compiled in the matlab version: upConv has a bug for
    # even-length kernels when using the reflect1, extend, or repeat edge-handlers
    if ((edge_type in ["reflect1", "extend", "repeat"]) and
            (filt.shape[0] % 2 == 0 or filt.shape[1] % 2 == 0)):
        if filt.shape[1] == 1:
            filt = np.append(filt, 0.0)
            filt = np.reshape(filt, (len(filt), 1))
        elif filt.shape[0] == 1:
            filt = np.append(filt, 0.0)
            filt = np.reshape(filt, (1, len(filt)))
        else:
            raise Exception('Even sized 2D filters not yet supported by upConv.')

    if stop is None:
        stop = [imshape_d * step_d for imshape_d, step_d in zip(image.shape, step)]

    result = np.zeros((stop[1], stop[0]))

    temp = np.zeros((filt.shape[1], filt.shape[0]))

    if edge_type == 'circular':
        lib.internal_wrap_expand(image.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                                 filt.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                                 filt.shape[1], filt.shape[0], start[1],
                                 step[1], stop[1], start[0], step[0], stop[0],
                                 result.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                                 stop[1], stop[0])
    else:
        lib.internal_expand(image.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                            filt.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                            temp.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                            filt.shape[1], filt.shape[0], start[1], step[1],
                            stop[1], start[0], step[0], stop[0],
                            result.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                            stop[1], stop[0], edge_type.encode('ascii'))
    result = np.reshape(result, stop)

    return result


def pointOp(image, lut, origin, increment, warnings=False):
    """Apply a point operation, specified by lookup table `lut`, to `image`

    This function is very fast and allows extrapolation beyond the lookup table domain.  The
    drawbacks are that the lookup table must be equi-spaced, and the interpolation is linear.

    Arguments
    ---------
    image : `array_like`
        1d or 2d array
    lut : `array_like`
        a row or column vector, assumed to contain (equi-spaced) samples of the function.
    origin : `float`
        specifies the abscissa associated with the first sample
    increment : `float`
        specifies the spacing between samples.
    warnings : `bool`
        whether to print a warning whenever the lookup table is extrapolated

    """
    result = np.empty_like(image)
    # this way we can use python booleans when calling
    if warnings:
        warnings = 1
    else:
        warnings = 0
    lib.internal_pointop(image.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                         result.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                         image.shape[0] * image.shape[1],
                         lut.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                         lut.shape[0],
                         ctypes.c_double(origin),
                         ctypes.c_double(increment), warnings)

    return np.asarray(result)



import unittest
import tarfile
import tqdm
import math
import requests

import numpy as np
import matplotlib.pyplot as plt

import pyrtools as pt
from pyrtools.pyramids.pyramid import Pyramid

import scipy.io
import os
import os.path as op
matfiles_path = op.join(op.dirname(op.realpath(__file__)), 'matFiles')
test_data_path = op.join(op.dirname(op.realpath(__file__)), '..', 'DATA')

# TODO:
# - create class upConvTests(unittest.TestCase):
# - expand class corrDnTests(unittest.TestCase):
# - expand class blurDnTests(unittest.TestCase):
# - clean up histo function and then run histoTests and entropy2Tests

class corrDnTests(unittest.TestCase):
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'corrDn1.mat'))
        mres = matPyr['res']
        ramp = pt.synthetic_images.ramp(20)
        res = pt.corrDn(ramp, pt.named_filter('binom5'))
        self.assertTrue(pt.compareRecon(mres, res))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'corrDn2.mat'))
        mres = matPyr['res']
        ramp = pt.synthetic_images.ramp(20)
        res = pt.corrDn(ramp, pt.named_filter('qmf13'))
        self.assertTrue(pt.compareRecon(mres, res))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'corrDn3.mat'))
        mres = matPyr['res']
        ramp = pt.synthetic_images.ramp(20)
        res = pt.corrDn(ramp, pt.named_filter('qmf16'))
        self.assertTrue(pt.compareRecon(mres, res))

class blurTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blur0.mat'))
        ramp = pt.synthetic_images.ramp(20)
        res = pt.blur(ramp)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blur1.mat'))
        ramp = pt.synthetic_images.ramp(20)
        res = pt.blur(ramp, n_levels=3)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blur2.mat'))
        ramp = pt.synthetic_images.ramp(20)
        res = pt.blur(ramp, n_levels=3, filt=pt.named_filter('qmf5'))
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blur3.mat'))
        ramp = pt.synthetic_images.ramp((20,30))
        res = pt.blur(ramp)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blur4.mat'))
        ramp = pt.synthetic_images.ramp((20,30))
        res = pt.blur(ramp, n_levels=3)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test5(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blur5.mat'))
        ramp = pt.synthetic_images.ramp((20,30))
        res = pt.blur(ramp, n_levels=3, filt=pt.named_filter('qmf5'))
        self.assertTrue(pt.compareRecon(matPyr['res'], res))


class blurDnTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blurDn0.mat'))
        pyRamp = pt.synthetic_images.ramp((20,20))
        res = pt.blurDn(pyRamp)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blurDn1.mat'))
        pyRamp = pt.synthetic_images.ramp((256,256))
        res = pt.blurDn(pyRamp)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blurDn2.mat'))
        pyRamp = pt.synthetic_images.ramp((256,128))
        res = pt.blurDn(pyRamp)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blurDn3.mat'))
        pyRamp = pt.synthetic_images.ramp((128,256))
        res = pt.blurDn(pyRamp)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blurDn4.mat'))
        pyRamp = pt.synthetic_images.ramp((200, 100))
        res = pt.blurDn(pyRamp)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test5(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blurDn5.mat'))
        pyRamp = pt.synthetic_images.ramp((100, 200))
        res = pt.blurDn(pyRamp)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test6(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blurDn6.mat'))
        pyRamp = pt.synthetic_images.ramp((1, 256)).T
        res = pt.blurDn(pyRamp)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test7(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blurDn7.mat'))
        pyRamp = pt.synthetic_images.ramp((1, 256))
        res = pt.blurDn(pyRamp)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    # def test8(self):#  need a 2D filter
    #    matPyr = scipy.io.loadmat(op.join(matfiles_path, 'blurDn8.mat'))
    #    pyRamp = pt.synthetic_images.ramp((256, 256))
    #    res = pt.blurDn(pyRamp, filt=2dfilt)
    #    self.assertTrue(pt.compareRecon(matPyr['res'], res))

class upBlurTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur0.mat'))
        mres = matPyr['res']
        im = pt.synthetic_images.ramp((1,20))
        res = pt.upBlur(im)
        self.assertTrue(pt.compareRecon(mres, res))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur1.mat'))
        mres = matPyr['res']
        im = pt.synthetic_images.ramp((1,20))
        res = pt.upBlur(im.T)
        self.assertTrue(pt.compareRecon(mres, res))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur2.mat'))
        mres = matPyr['res']
        im = pt.synthetic_images.ramp(20)
        res = pt.upBlur(im)
        self.assertTrue(pt.compareRecon(mres, res))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur3.mat'))
        mres = matPyr['res']
        im = pt.synthetic_images.ramp((1,20))
        res = pt.upBlur(im, n_levels=3)
        self.assertTrue(pt.compareRecon(mres, res))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur4.mat'))
        mres = matPyr['res']
        im = pt.synthetic_images.ramp((1,20))
        res = pt.upBlur(im.T, n_levels=3)
        self.assertTrue(pt.compareRecon(mres, res))
    def test5(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur5.mat'))
        mres = matPyr['res']
        im = pt.synthetic_images.ramp(20)
        res = pt.upBlur(im, n_levels=3)
        self.assertTrue(pt.compareRecon(mres, res))
    def test6(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur6.mat'))
        mres = matPyr['res']
        im = pt.synthetic_images.ramp((1,20))
        filt = pt.named_filter('qmf9')
        res = pt.upBlur(im, n_levels=3, filt=filt)
        self.assertTrue(pt.compareRecon(mres, res))
    #def test7(self):   # fails in matlab and python because of dim mismatch
    #    matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur7.mat'))
    #    mres = matPyr['res']
    #    im = pt.synthetic_images.ramp((1,20))
    #    filt = pt.named_filter('qmf9')
    #    res = pt.upBlur(im, 3, filt.T)
    #    self.assertTrue(pt.compareRecon(mres, res))
    def test8(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur8.mat'))
        mres = matPyr['res']
        im = pt.synthetic_images.ramp((1,20))
        filt = pt.named_filter('qmf9')
        res = pt.upBlur(im.T, n_levels=3, filt=filt)
        self.assertTrue(pt.compareRecon(mres, res))
    #def test9(self):  # fails in matlab and python because of dim mismatch
    #    matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur6.mat'))
    #    mres = matPyr['res']
    #    im = pt.synthetic_images.ramp((1,20))
    #    filt = pt.named_filter('qmf9')
    #    res = pt.upBlur(im.T, 3, filt.T)
    #    self.assertTrue(pt.compareRecon(mres, res))
    def test10(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur10.mat'))
        mres = matPyr['res']
        im = pt.synthetic_images.ramp(20)
        filt = pt.synthetic_images.disk(3)
        res = pt.upBlur(im, n_levels=3, filt=filt)
        self.assertTrue(pt.compareRecon(mres, res))
    def test11(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur11.mat'))
        mres = matPyr['res']
        im = pt.synthetic_images.ramp((20,10))
        filt = pt.synthetic_images.disk((5,3))
        res = pt.upBlur(im, n_levels=3, filt=filt)
        self.assertTrue(pt.compareRecon(mres, res))
    def test12(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'upBlur12.mat'))
        mres = matPyr['res']
        im = pt.synthetic_images.ramp((10,20))
        filt = pt.synthetic_images.disk((3,5))
        res = pt.upBlur(im, n_levels=3, filt=filt)
        self.assertTrue(pt.compareRecon(mres, res))

class pointOpTests(unittest.TestCase):
    def test1(self):
        matImg = scipy.io.loadmat(op.join(matfiles_path, 'pointOp1.mat'))
        img = pt.synthetic_images.ramp((200,200))
        filt = np.asarray([0.2, 0.5, 1.0, 0.4, 0.1]);
        #foo = pointOp(200, 200, img, 5, filt, 0, 1, 0);
        foo = pt.pointOp(img, filt, 0, 1);
        foo = np.reshape(foo,(200,200))
        np.testing.assert_allclose(matImg['foo'], foo)

class maxPyrHeightTests(unittest.TestCase):
    def test1(self):
        self.assertTrue(pt.pyramids.max_pyr_height((1,10),(3,4)) == 0)
    def test2(self):
        self.assertTrue(pt.pyramids.max_pyr_height((10,1),(3,4)) == 0)
    def test3(self):
        self.assertTrue(pt.pyramids.max_pyr_height((10,10),(1,4)) == 2)
    def test4(self):
        self.assertTrue(pt.pyramids.max_pyr_height((10,10),(3,1)) == 2)
    def test5(self):
        self.assertTrue(pt.pyramids.max_pyr_height((10,10),(3,4)) == 2)
    def test6(self):
        self.assertTrue(pt.pyramids.max_pyr_height((20,10),(5,1)) == 2)
    def test7(self):
        self.assertTrue(pt.pyramids.max_pyr_height((10,20),(5,1)) == 2)
    def test8(self):
        self.assertTrue(pt.pyramids.max_pyr_height((20,10),(1,5)) == 2)
    def test9(self):
        self.assertTrue(pt.pyramids.max_pyr_height((10,20),(1,5)) == 2)
    def test10(self):
        self.assertTrue(pt.pyramids.max_pyr_height((256,1),(1,5)) == 6)
    def test11(self):
        self.assertTrue(pt.pyramids.max_pyr_height((256,1),(5,1)) == 6)
    def test12(self):
        self.assertTrue(pt.pyramids.max_pyr_height((1,256),(1,5)) == 6)
    def test13(self):
        self.assertTrue(pt.pyramids.max_pyr_height((1,256),(5,1)) == 6)

class binomialFilterTests(unittest.TestCase):
    def test1(self):
        target = np.asarray([[0.5],[0.5]])
        #target = target / np.sqrt(np.sum(target ** 2))
        self.assertTrue((pt.binomial_filter(2) == target).all() )
    def test2(self):
        target = np.asarray([[0.25], [0.5], [0.25]])
        #target = target / np.sqrt(np.sum(target ** 2))
        self.assertTrue((pt.binomial_filter(3) == target).all())
    def test3(self):
        target = np.asarray([[0.0625], [0.25], [0.3750], [0.25], [0.0625]])
        #target = target / np.sqrt(np.sum(target ** 2))
        self.assertTrue((pt.binomial_filter(5) == target).all())

class GpyrTests(unittest.TestCase):
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildGpyr1.mat'))
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyPyr = pt.pyramids.GaussianPyramid(img)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildGpyr2row.mat'))
        img = np.asarray(list(range(256))).astype(float)
        img = img.reshape(1, 256)
        pyPyr = pt.pyramids.GaussianPyramid(img)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildGpyr2col.mat'))
        img = np.asarray(list(range(256))).astype(float)
        img = img.reshape(256, 1)
        pyPyr = pt.pyramids.GaussianPyramid(img)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildGpyr3.mat'))
        img = pt.synthetic_images.ramp(10)
        pyPyr = pt.pyramids.GaussianPyramid(img)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test5(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildGpyr4.mat'))
        img = pt.synthetic_images.ramp((10,20))
        pyPyr = pt.pyramids.GaussianPyramid(img)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test6(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildGpyr5.mat'))
        img = pt.synthetic_images.ramp((20, 10))
        pyPyr = pt.pyramids.GaussianPyramid(img)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))

class LpyrTests(unittest.TestCase):
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr1.mat'))
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyPyr = pt.pyramids.LaplacianPyramid(img)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test2(self):
        matPyr2 = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr2.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.LaplacianPyramid(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr2['pyr'], pyPyr))
    def test3(self):
        matPyr2 = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr3.mat'))
        pyRamp = pt.synthetic_images.ramp((100,200))
        pyPyr = pt.pyramids.LaplacianPyramid(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr2['pyr'], pyPyr))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr4.mat'))
        pyRamp = pt.synthetic_images.ramp(200,100)
        pyPyr = pt.pyramids.LaplacianPyramid(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test5(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr5.mat'))
        pyRamp = np.asarray(list(range(200))).reshape(1, 200)
        pyPyr = pt.pyramids.LaplacianPyramid(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test5bis(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr5.mat'))
        pyRamp = np.asarray(list(range(200)))
        pyPyr = pt.pyramids.LaplacianPyramid(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test6(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr6.mat'))
        pyRamp = np.asarray(list(range(200)))
        pyPyr = pt.pyramids.LaplacianPyramid(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test7(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr7.mat'))
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyPyr = pt.pyramids.LaplacianPyramid(img)
        recon = pyPyr.recon_pyr()
        np.testing.assert_allclose(matPyr['recon'], recon)
    def test8(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr8.mat'))
        pyRamp = pt.synthetic_images.ramp(200)
        pyPyr = pt.pyramids.LaplacianPyramid(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue((matPyr['recon'] == recon).all())
    def test9(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr9.mat'))
        pyRamp = pt.synthetic_images.ramp((200,100))
        pyPyr = pt.pyramids.LaplacianPyramid(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue((matPyr['recon'] == recon).all())
    def test10(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr10.mat'))
        pyRamp = pt.synthetic_images.ramp((100,200))
        pyPyr = pt.pyramids.LaplacianPyramid(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue((matPyr['recon'] == recon).all())
    def test11(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr11.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.LaplacianPyramid(pyRamp)
        recon = pyPyr.recon_pyr(levels=[1])
        np.testing.assert_allclose(matPyr['recon'], recon, atol=1e-6)
    def test12(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildLpyr12.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.LaplacianPyramid(pyRamp)
        recon = pyPyr.recon_pyr(levels=[0, 2, 4])
        np.testing.assert_allclose(matPyr['recon'], recon, atol=1e-6)

class WpyrTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr0.mat'))
        pyRamp = pt.synthetic_images.ramp((20,20))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr1.mat'))
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyPyr = pt.pyramids.WaveletPyramid(img)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr2.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr3.mat'))
        pyRamp = pt.synthetic_images.ramp((100,200))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr4.mat'))
        pyRamp = pt.synthetic_images.ramp((200,100))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test5(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr5.mat'))
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyPyr = pt.pyramids.WaveletPyramid(img)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test6(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr6.mat'))
        pyRamp = pt.synthetic_images.ramp((256,128))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test7(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr7.mat'))
        pyRamp = pt.synthetic_images.ramp((128,256))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test8(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr8.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test9(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr9.mat'))
        pyRamp = pt.synthetic_images.ramp((200,100))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test10(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr10.mat'))
        pyRamp = pt.synthetic_images.ramp((100,200))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test11(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr11.mat'))
        pyRamp = pt.synthetic_images.ramp((256,256))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        recon = pyPyr.recon_pyr('qmf9', 'reflect1', [0])
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test12(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr12.mat'))
        pyRamp = pt.synthetic_images.ramp((256,256))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        recon = pyPyr.recon_pyr('qmf9', 'reflect1', [0,2,4])
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test13(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr13.mat'))
        pyRamp = pt.synthetic_images.ramp((256,256))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        recon = pyPyr.recon_pyr('qmf9', 'reflect1', [0,2,4], [1])
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test14(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr14.mat'))
        pyRamp = pt.synthetic_images.ramp((256,256))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        recon = pyPyr.recon_pyr('qmf8')
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test15(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr15.mat'))
        pyRamp = pt.synthetic_images.ramp((256,128))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        recon = pyPyr.recon_pyr('qmf8')
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test16(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr16.mat'))
        pyRamp = pt.synthetic_images.ramp((128,256))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        recon = pyPyr.recon_pyr('qmf8')
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test17(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr17.mat'))
        pyRamp = pt.synthetic_images.ramp((1,200))
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test18(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildWpyr18.mat'))
        pyRamp = pt.synthetic_images.ramp((1,200)).T
        pyPyr = pt.pyramids.WaveletPyramid(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def testRecon1(self):
        np.random.seed(0)
        im = pt.synthetic_images.pink_noise((1, 64))
        pyr = pt.pyramids.WaveletPyramid(im)
        res = pyr.recon_pyr()
        self.assertTrue(np.allclose(res, im, atol=5e-3))
    def testRecon2(self):
        np.random.seed(0)
        im = pt.synthetic_images.pink_noise((64, 1))
        pyr = pt.pyramids.WaveletPyramid(im)
        res = pyr.recon_pyr()
        self.assertTrue(np.allclose(res, im, atol=5e-3))
    def testRecon3(self):
        im = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyr = pt.pyramids.WaveletPyramid(im)
        res = pyr.recon_pyr()
        # print("\n%s\n" % np.max(res-im))
        self.assertTrue(np.allclose(res, im, atol=1))
    def testRecon4(self):
        im = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyr = pt.pyramids.WaveletPyramid(im, edge_type='circular')
        res = pyr.recon_pyr()
        # print("\n%s\n" % np.max(res-im))
        self.assertTrue(np.allclose(res, im, atol=1))
    def testRecon5(self):
        im = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyr = pt.pyramids.WaveletPyramid(im, filter_name='daub2', edge_type='circular')
        res = pyr.recon_pyr()
        # print("\n%s\n" % np.max(res-im))
        self.assertTrue(np.allclose(res, im, atol=5e-3))
    def testRecon6(self):
        im = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyr = pt.pyramids.WaveletPyramid(im, filter_name='qmf13', edge_type='circular')
        res = pyr.recon_pyr()
        # print("\n%s\n" % np.max(res-im))
        self.assertTrue(np.allclose(res, im, atol=2))
    def testRecon7(self):
        im = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyr = pt.pyramids.WaveletPyramid(im, filter_name='haar', edge_type='circular')
        res = pyr.recon_pyr()
        self.assertTrue(np.allclose(res, im))

class spFilterTests(unittest.TestCase):
    def test1(self):
        matFilt0 = scipy.io.loadmat(op.join(matfiles_path, 'sp0Filters.mat'))
        pySP0filt = pt.steerable_filters('sp0_filters')
        tmpKeys = []
        for key in list(matFilt0.keys()):
            if "_" not in key:
                tmpKeys.append(key)
        self.assertTrue(set(tmpKeys) == set(pySP0filt.keys()))
        for key in tmpKeys:
            self.assertTrue((matFilt0[key] == pySP0filt[key]).all())

    def test2(self):
        matFilt1 = scipy.io.loadmat(op.join(matfiles_path, 'sp1Filters.mat'))
        pySP1filt = pt.steerable_filters('sp1_filters')
        tmpKeys = []
        for key in list(matFilt1.keys()):
            if "_" not in key:
                tmpKeys.append(key)
        self.assertTrue(set(tmpKeys) == set(pySP1filt.keys()))
        for key in tmpKeys:
            self.assertTrue((matFilt1[key] == pySP1filt[key]).all())

    def test3(self):
        matFilt3 = scipy.io.loadmat(op.join(matfiles_path, 'sp3Filters.mat'))
        pySP3filt = pt.steerable_filters('sp3_filters')
        tmpKeys = []
        for key in list(matFilt3.keys()):
            if "_" not in key:
                tmpKeys.append(key)
        self.assertTrue(set(tmpKeys) == set(pySP3filt.keys()))
        for key in tmpKeys:
            self.assertTrue((matFilt3[key] == pySP3filt[key]).all())

    def test4(self):
        matFilt5 = scipy.io.loadmat(op.join(matfiles_path, 'sp5Filters.mat'))
        pySP5filt = pt.steerable_filters('sp5_filters')
        tmpKeys = []
        for key in list(matFilt5.keys()):
            if "_" not in key:
                tmpKeys.append(key)
        self.assertTrue(set(tmpKeys) == set(pySP5filt.keys()))
        for key in tmpKeys:
            self.assertTrue((matFilt5[key] == pySP5filt[key]).all())

class SteerablePyramidSpaceTests(unittest.TestCase):
    def test00(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr00.mat'))
        pyRamp = pt.synthetic_images.ramp((20,20))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr1.mat'))
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyPyr = pt.pyramids.SteerablePyramidSpace(img)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr2.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr3.mat'))
        pyRamp = pt.synthetic_images.ramp((100,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr4.mat'))
        pyRamp = pt.synthetic_images.ramp((200,100))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr0.mat'))
        pyRamp = pt.synthetic_images.ramp(20)
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test5(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr5.mat'))
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyPyr = pt.pyramids.SteerablePyramidSpace(img)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test6(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr6.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test7(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr7.mat'))
        pyRamp = pt.synthetic_images.ramp((256,128))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test8(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr8.mat'))
        pyRamp = pt.synthetic_images.ramp((128,256))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test9(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr9.mat'))
        pyRamp = pt.synthetic_images.ramp((200,100))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test10(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr10.mat'))
        pyRamp = pt.synthetic_images.ramp((100,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test11(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr11.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp)
        recon = pyPyr.recon_pyr(1, 'reflect1', [0])
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test12(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr12.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp)
        recon = pyPyr.recon_pyr(1, 'reflect1',
                                ['residual_highpass', 1, 3])
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test13(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr13.mat'))
        pyRamp = pt.synthetic_images.ramp((20,20))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp, height=1, order=0)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test14(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr14.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp, height=3, order=0)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test15(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr15.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp, height=1, order=1)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test16(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr16.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp, height=3, order=1)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test17(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr17.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp, height=1, order=3)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test18(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr18.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp, height=3, order=3)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test19(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr19.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp, height=1, order=5)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test20(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr20.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidSpace(pyRamp, height=3, order=5)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test21(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr21.mat'))
        texture = scipy.io.loadmat(op.join(matfiles_path, 'im04-1.mat'))['res'][0:256,0:256]
        pyPyr = pt.pyramids.SteerablePyramidSpace(texture, height=3, order=0)
        recon = pyPyr.recon_pyr(0);
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test22(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr22.mat'))
        texture = scipy.io.loadmat(op.join(matfiles_path, 'im04-1.mat'))['res'][0:256,0:256]
        pyPyr = pt.pyramids.SteerablePyramidSpace(texture, height=3, order=1)
        recon = pyPyr.recon_pyr(1);
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test23(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr23.mat'))
        texture = scipy.io.loadmat(op.join(matfiles_path, 'im04-1.mat'))['res'][0:256,0:256]
        pyPyr = pt.pyramids.SteerablePyramidSpace(texture, height=3, order=3)
        recon = pyPyr.recon_pyr(3);
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test24(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr24.mat'))
        texture = scipy.io.loadmat(op.join(matfiles_path, 'im04-1.mat'))['res'][0:256,0:256]
        pyPyr = pt.pyramids.SteerablePyramidSpace(texture, height=3, order=5)
        recon = pyPyr.recon_pyr(5);
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test25(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr25.mat'))
        texture = scipy.io.loadmat(op.join(matfiles_path, 'im04-1.mat'))['res'][0:256,0:256]
        pyPyr = pt.pyramids.SteerablePyramidSpace(texture, height=3, order=5)
        recon = pyPyr.recon_pyr(5,'reflect1', ['residual_highpass', 0, 1], [0]);
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test26(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr26.mat'))
        texture = scipy.io.loadmat(op.join(matfiles_path, 'im04-1.mat'))['res'][0:128,0:256]
        pyPyr = pt.pyramids.SteerablePyramidSpace(texture, height=3, order=0)
        recon = pyPyr.recon_pyr(0,'reflect1', ['residual_highpass', 0, 1], [0]);
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test27(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr27.mat'))
        texture = scipy.io.loadmat(op.join(matfiles_path, 'im04-1.mat'))['res'][0:128,0:256]
        pyPyr = pt.pyramids.SteerablePyramidSpace(texture, height=3, order=1)
        recon = pyPyr.recon_pyr(1,'reflect1', ['residual_highpass', 0, 1], [0]);
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test28(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr28.mat'))
        texture = scipy.io.loadmat(op.join(matfiles_path, 'im04-1.mat'))['res'][0:128,0:256]
        pyPyr = pt.pyramids.SteerablePyramidSpace(texture, height=3, order=3)
        recon = pyPyr.recon_pyr(3,'reflect1', ['residual_highpass', 0, 1], [0]);
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test29(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSpyr29.mat'))
        texture = scipy.io.loadmat(op.join(matfiles_path, 'im04-1.mat'))['res'][0:128,0:256]
        pyPyr = pt.pyramids.SteerablePyramidSpace(texture, height=3, order=5)
        recon = pyPyr.recon_pyr(5,'reflect1', ['residual_highpass', 0, 1], [0]);
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))

class SteerablePyramidFreqpyrTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr0.mat'))
        pyRamp = pt.synthetic_images.ramp((20,20))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr1.mat'))
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyPyr = pt.pyramids.SteerablePyramidFreq(img)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr2.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr3.mat'))
        pyRamp = pt.synthetic_images.ramp((100,200))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr4.mat'))
        pyRamp = pt.synthetic_images.ramp((200,100))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test5(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr5.mat'))
        pyRamp = pt.synthetic_images.ramp((20,20))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test6(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr6.mat'))
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyPyr = pt.pyramids.SteerablePyramidFreq(img)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test7(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr7.mat'))
        pyRamp = pt.synthetic_images.ramp((256,128))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test8(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr8.mat'))
        pyRamp = pt.synthetic_images.ramp((128,256))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test9(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr9.mat'))
        pyRamp = pt.synthetic_images.ramp((200,100))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test10(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr10.mat'))
        pyRamp = pt.synthetic_images.ramp((100,200))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test11(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr11.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp)
        recon = pyPyr.recon_pyr(['residual_highpass'])
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test12(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr12.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp)
        recon = pyPyr.recon_pyr(['residual_highpass', 1, 3])
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test13(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSFpyr13.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp)
        recon = pyPyr.recon_pyr(['residual_highpass', 1, 3], [1])
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def testRecon1(self):
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyr = pt.pyramids.SteerablePyramidFreq(img)
        recon = pyr.recon_pyr()
        self.assertTrue(np.allclose(img, recon, atol=5e-3))
    # we initially had a bug with odd orientations, since fixed
    def testRecon2(self):
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyr = pt.pyramids.SteerablePyramidFreq(img, order=2)
        recon = pyr.recon_pyr()
        self.assertTrue(np.allclose(img, recon, atol=5e-3))

class SteerablePyramidComplexTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSCFpyr0.mat'))
        pyRamp = pt.synthetic_images.ramp((20,20))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp, is_complex=True)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSCFpyr1.mat'))
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyPyr = pt.pyramids.SteerablePyramidFreq(img, is_complex=True)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSCFpyr2.mat'))
        pyRamp = pt.synthetic_images.ramp((200,200))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp, is_complex=True)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSCFpyr3.mat'))
        pyRamp = pt.synthetic_images.ramp((100,200))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp, is_complex=True)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSCFpyr4.mat'))
        pyRamp = pt.synthetic_images.ramp((200,100))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp, is_complex=True)
        self.assertTrue(pt.comparePyr(matPyr['pyr'], pyPyr))
    def test5(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSCFpyr5.mat'))
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyPyr = pt.pyramids.SteerablePyramidFreq(img, is_complex=True)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test6(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSCFpyr6.mat'))
        pyRamp = pt.synthetic_images.ramp((256,128))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp, is_complex=True)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test7(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSCFpyr7.mat'))
        pyRamp = pt.synthetic_images.ramp((128,256))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp, is_complex=True)
        recon = pyPyr.recon_pyr()
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test10(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSCFpyr10.mat'))
        pyRamp = pt.synthetic_images.ramp((256,256))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp, is_complex=True)
        recon = pyPyr.recon_pyr(['residual_highpass'])
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test11(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSCFpyr11.mat'))
        pyRamp = pt.synthetic_images.ramp((256,256))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp, is_complex=True)
        recon = pyPyr.recon_pyr(['residual_highpass', 1, 3])
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def test12(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'buildSCFpyr12.mat'))
        pyRamp = pt.synthetic_images.ramp((256,256))
        pyPyr = pt.pyramids.SteerablePyramidFreq(pyRamp, is_complex=True)
        recon = pyPyr.recon_pyr(['residual_highpass', 1, 3], [1])
        self.assertTrue(pt.compareRecon(matPyr['recon'], recon))
    def testRecon1(self):
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyr = pt.pyramids.SteerablePyramidFreq(img, is_complex=True)
        recon = pyr.recon_pyr()
        self.assertTrue(np.allclose(img, recon, atol=5e-3))
    # we initially had a bug with odd orientations, since fixed
    def testRecon2(self):
        img = plt.imread(op.join(test_data_path, 'lenna-256x256.tif'))
        pyr = pt.pyramids.SteerablePyramidFreq(img, order=2, is_complex=True)
        recon = pyr.recon_pyr()
        self.assertTrue(np.allclose(img, recon, atol=5e-3))

class mkAngularSineTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkAngularSine0.mat'))
        res = pt.synthetic_images.angular_sine(20)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkAngularSine1.mat'))
        res = pt.synthetic_images.angular_sine(20, 5)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkAngularSine2.mat'))
        res = pt.synthetic_images.angular_sine(20, 5, 3)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkAngularSine3.mat'))
        res = pt.synthetic_images.angular_sine(20, 5, 3, 2)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkAngularSine4.mat'))
        res = pt.synthetic_images.angular_sine(20, 5, 3, 2, (2,2))
        self.assertTrue(pt.compareRecon(matPyr['res'], res))

class mkGaussianTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkGaussian0.mat'))
        res = pt.synthetic_images.gaussian(20)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkGaussian1.mat'))
        res = pt.synthetic_images.gaussian(20, (2,3))
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkGaussian2.mat'))
        res = pt.synthetic_images.gaussian(20, [[-1, 0], [0, 1]])
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkGaussian3.mat'))
        res = pt.synthetic_images.gaussian(10, [[-1, 0], [0, 1]])
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkGaussian4.mat'))
        res = pt.synthetic_images.gaussian(20, [[2, 0], [0, 1]])
        self.assertTrue(pt.compareRecon(matPyr['res'], res))

class mkDiscTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkDisc0.mat'))
        res = pt.synthetic_images.disk(20)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkDisc1.mat'))
        res = pt.synthetic_images.disk(20, 8)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkDisc2.mat'))
        res = pt.synthetic_images.disk(20, 8, (0,0))
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkDisc3.mat'))
        res = pt.synthetic_images.disk(20, 8, (0,0), 5)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkDisc4.mat'))
        res = pt.synthetic_images.disk(20, 8, (0,0), 5, (0.75, 0.25))
        self.assertTrue(pt.compareRecon(matPyr['res'], res))

class mkSineTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSine0.mat'))
        res = pt.synthetic_images.sine(20, period=5.5)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSine1.mat'))
        res = pt.synthetic_images.sine(20, period=5.5, direction=2)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSine2.mat'))
        res = pt.synthetic_images.sine(20, period=5.5, direction=2, amplitude=3)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSine3.mat'))
        res = pt.synthetic_images.sine(20, period=5.5, direction=2, amplitude=3, phase=5)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSine4.mat'))
        res = pt.synthetic_images.sine(20, period=5.5, direction=2, amplitude=3, phase=5, origin=[4,5])
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test5(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSine5.mat'))
        res = pt.synthetic_images.sine(20, frequency=[1,2])
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test6(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSine6.mat'))
        res = pt.synthetic_images.sine(20, frequency=[1,2], amplitude=3)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test7(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSine7.mat'))
        res = pt.synthetic_images.sine(20, frequency=[1,2], amplitude=3, phase=2)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test8(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSine8.mat'))
        res = pt.synthetic_images.sine(20, frequency=[1,2], amplitude=3, phase=2, origin=[5,4])
        self.assertTrue(pt.compareRecon(matPyr['res'], res))

class mkZonePlateTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkZonePlate0.mat'))
        res = pt.synthetic_images.zone_plate(20)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkZonePlate1.mat'))
        res = pt.synthetic_images.zone_plate(20, 4)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkZonePlate2.mat'))
        res = pt.synthetic_images.zone_plate(20, 4, 3)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))

class mkSquareTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSquare0.mat'))
        res = pt.synthetic_images.square_wave(20, period=5.5)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSquare1.mat'))
        res = pt.synthetic_images.square_wave(20, period=5.5, direction=3)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSquare2.mat'))
        res = pt.synthetic_images.square_wave(20, period=5.5, direction=3, amplitude=5.1)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test3(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSquare3.mat'))
        res = pt.synthetic_images.square_wave(20, period=5.5, direction=3, amplitude=5.1, phase=-1)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test4(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSquare4.mat'))
        res = pt.synthetic_images.square_wave(20, period=5.5, direction=3, amplitude=5.1, phase=-1, origin=(2,3))
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test5(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSquare5.mat'))
        res = pt.synthetic_images.square_wave(20, period=5.5, direction=3, amplitude=5.1, phase=-1, origin=(2,3), twidth=.25)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test6(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSquare6.mat'))
        res = pt.synthetic_images.square_wave(20, frequency=(1,2))
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test7(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSquare7.mat'))
        res = pt.synthetic_images.square_wave(20, frequency=(1,2), amplitude=3.2)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test8(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSquare8.mat'))
        res = pt.synthetic_images.square_wave(20, frequency=(1,2), amplitude=3.2, phase=-2)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test9(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSquare9.mat'))
        res = pt.synthetic_images.square_wave(20, frequency=(1,2), amplitude=3.2, phase=-2, origin=(2,3))
        self.assertTrue(pt.compareRecon(matPyr['res'], res))
    def test10(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'mkSquare10.mat'))
        res = pt.synthetic_images.square_wave(20, frequency=(1,2), amplitude=3.2, phase=-2, origin=(2,3), twidth=.55)
        self.assertTrue(pt.compareRecon(matPyr['res'], res))

class polarAngleTests(unittest.TestCase):
    def test0(self):
        ang = pt.synthetic_images.polar_angle(100, direction='clockwise')
        idx = np.argmin((ang - np.pi/2)**2)
        # check that pi/2 is in the bottom half of the image
        self.assertTrue(np.unravel_index(idx, (100, 100))[0] > 50)
        idx = np.argmin((ang + np.pi/2)**2)
        # check that -pi/2 is in the top half of the image
        self.assertTrue(np.unravel_index(idx, (100, 100))[0] < 50)
    def test1(self):
        ang = pt.synthetic_images.polar_angle(100, direction='counter-clockwise')
        idx = np.argmin((ang - np.pi/2)**2)
        # check that pi/2 is in the top half of the image
        self.assertTrue(np.unravel_index(idx, (100, 100))[0] < 50)
        idx = np.argmin((ang + np.pi/2)**2)
        # check that -pi/2 is in the bottom half of the image
        self.assertTrue(np.unravel_index(idx, (100, 100))[0] > 50)
    def test2(self):
        with self.assertRaises(ValueError):
            pt.synthetic_images.polar_angle(100, direction='-clockwise')

# TODO

# python version of histo
# adding 0.7 to ramp to nullify rounding differences between Python and Matlab

# class histoTests(unittest.TestCase):
#     def test0(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'histo0.mat'))
#         (N,X) = pt.matlab_histo(pt.synthetic_images.ramp(10) + 0.7)
#         # X will not be the same because matlab returns centers and
#         #   python return edges
#         #self.assertTrue(pt.compareRecon(matPyr['X'], X))
#         #self.assertTrue(pt.compareRecon(matPyr['N'], N))
#         self.assertTrue((matPyr['N'] == N).all())

# FIX: why does matlab version return N+1 bins??
#    def test1(self):
#        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'histo1.mat'))
#        (N,X) = pt.histo(pt.synthetic_images.ramp(10) + 0.7, 26)
#        # X will not be the same because matlab returns centers and
#        #   python return edges
#        #self.assertTrue(pt.compareRecon(matPyr['X'], X))
#        #self.assertTrue(pt.compareRecon(matPyr['N'], N))
#        self.assertTrue(matPyr['N'] == N)
#    def test2(self):
#        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'histo2.mat'))
#        (N,X) = pt.histo(pt.synthetic_images.ramp(10) + 0.7, -1.15)
#        # X will not be the same because matlab returns centers and
#        #   python return edges
#        #self.assertTrue(pt.compareRecon(matPyr['X'], X))
#        self.assertTrue(pt.compareRecon(matPyr['N'], N))
#    def test3(self):
#        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'histo3.mat'))
#        (N,X) = pt.histo(pt.synthetic_images.ramp(10) + 0.7, 26, 3)
#        # X will not be the same because matlab returns centers and
#        #   python return edges
#        #self.assertTrue(pt.compareRecon(matPyr['X'], X))
#        self.assertTrue(pt.compareRecon(matPyr['N'], N))

# class entropy2Tests(unittest.TestCase):
#    # def test0(self):
#    #     matPyr = scipy.io.loadmat(op.join(matfiles_path, 'entropy2_0.mat'))
#    #     H = pt.entropy(pt.synthetic_images.ramp(10))
#    #     self.assertTrue(matPyr['H'] == H)
#
#    def test1(self):
#        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'entropy2_1.mat'))
#        H = pt.entropy(pt.synthetic_images.ramp(10), 1)
#        self.assertTrue(matPyr['H'] == H)

class ImageGradientTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'imGradient0.mat'))
        ramp = pt.synthetic_images.ramp(10)
        [dx,dy] = pt.image_gradient(ramp)
        dx = np.asarray(dx)
        dy = np.asarray(dy)
        self.assertTrue(pt.compareRecon(matPyr['res'][:,:,0], dx))
        self.assertTrue(pt.compareRecon(matPyr['res'][:,:,1], dy))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'imGradient1.mat'))
        ramp = pt.synthetic_images.ramp(10)
        [dx,dy] = pt.image_gradient(ramp, 'reflect1')
        dx = np.asarray(dx)
        dy = np.asarray(dy)
        self.assertTrue(pt.compareRecon(matPyr['res'][:,:,0], dx))
        self.assertTrue(pt.compareRecon(matPyr['res'][:,:,1], dy))

class skewTests(unittest.TestCase):
    def test0(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'skew2_0.mat'))
        # not sure why matPyr is [[ans]]???
        mres = matPyr['res'][0][0]
        disc = pt.synthetic_images.disk(10)
        res = pt.skew(disc)
        self.assertTrue(np.absolute(res - mres) <= np.power(10.0,-11))
    def test1(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'skew2_1.mat'))
        # not sure why matPyr is [[ans]]???
        mres = matPyr['res'][0][0]
        disc = pt.synthetic_images.disk(10)
        # using incorrect mean for better test
        mn = disc.mean() + 0.1
        res = pt.skew(disc, mn)
        self.assertTrue(np.absolute(res - mres) <= np.power(10.0,-11))
    def test2(self):
        matPyr = scipy.io.loadmat(op.join(matfiles_path, 'skew2_2.mat'))
        # not sure why matPyr is [[ans]]???
        mres = matPyr['res'][0][0]
        disc = pt.synthetic_images.disk(10)
        # using incorrect mean for better test
        mn = disc.mean() + 0.1
        v = pt.var(disc) + 0.1
        res = pt.skew(disc, mn, v)
        self.assertTrue(np.absolute(res - mres) <= np.power(10.0,-11))


class ProjectPolarTests(unittest.TestCase):
    def test0(self):
        # check that it runs, even though here the output will be nonsensical
        img = pt.synthetic_images.disk(256)
        proj_img = pt.project_polar_to_cartesian(img)
    def test1(self):
        # currently only works for square images
        img = pt.synthetic_images.disk((256, 512))
        with self.assertRaises(Exception):
            pt.project_polar_to_cartesian(img)


# class cconv2Tests(unittest.TestCase):
#     def test0(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'cconv2_0.mat'))
#         res = pt.cconv2(pt.synthetic_images.ramp(20), pt.synthetic_images.ramp(10))
#         self.assertTrue(pt.compareRecon(matPyr['res'], res))
#     def test1(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'cconv2_1.mat'))
#         res = pt.cconv2(pt.synthetic_images.ramp(10), pt.synthetic_images.ramp(20))
#         self.assertTrue(pt.compareRecon(matPyr['res'], res))
#     def test2(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'cconv2_2.mat'))
#         res = pt.cconv2(pt.synthetic_images.ramp(20), pt.synthetic_images.ramp(10), 3)
#         self.assertTrue(pt.compareRecon(matPyr['res'], res))
#     def test3(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'cconv2_3.mat'))
#         res = pt.cconv2(pt.synthetic_images.ramp(10), pt.synthetic_images.ramp(20), 3)
#         self.assertTrue(pt.compareRecon(matPyr['res'], res))
#     def test4(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'cconv2_4.mat'))
#         res = pt.cconv2(pt.synthetic_images.ramp((20,30)), pt.synthetic_images.ramp((10,20)))
#         self.assertTrue(pt.compareRecon(matPyr['res'], res))
#     def test5(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'cconv2_5.mat'))
#         res = pt.cconv2(pt.synthetic_images.ramp((10,20)), pt.synthetic_images.ramp((20,30)))
#         self.assertTrue(pt.compareRecon(matPyr['res'], res))
#     def test6(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'cconv2_6.mat'))
#         res = pt.cconv2(pt.synthetic_images.ramp((20,30)), pt.synthetic_images.ramp((10,20)), 5)
#         self.assertTrue(pt.compareRecon(matPyr['res'], res))
#     def test7(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'cconv2_7.mat'))
#         res = pt.cconv2(pt.synthetic_images.ramp((10,20)), pt.synthetic_images.ramp((20,30)), 5)
#         self.assertTrue(pt.compareRecon(matPyr['res'], res))

# class zconv2Tests(unittest.TestCase):
#     def test0(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'zconv2_0.mat'))
#         mres = matPyr['res']
#         ramp = pt.synthetic_images.ramp(10)
#         disc = pt.synthetic_images.disk(5)
#         res = pt.zconv2(ramp, disc)
#         self.assertTrue(pt.compareRecon(mres, res))
#     def test1(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'zconv2_1.mat'))
#         mres = matPyr['res']
#         ramp = pt.synthetic_images.ramp((10,20))
#         disc = pt.synthetic_images.disk((5,10))
#         res = pt.zconv2(ramp, disc)
#         self.assertTrue(pt.compareRecon(mres, res))
#     def test2(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'zconv2_2.mat'))
#         mres = matPyr['res']
#         ramp = pt.synthetic_images.ramp((20,10))
#         disc = pt.synthetic_images.disk((10,5))
#         res = pt.zconv2(ramp, disc)
#         self.assertTrue(pt.compareRecon(mres, res))
#     def test3(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'zconv2_3.mat'))
#         mres = matPyr['res']
#         ramp = pt.synthetic_images.ramp(10)
#         disc = pt.synthetic_images.disk(5)
#         res = pt.zconv2(ramp, disc, 3)
#         self.assertTrue(pt.compareRecon(mres, res))
#     def test4(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'zconv2_4.mat'))
#         mres = matPyr['res']
#         ramp = pt.synthetic_images.ramp((10,20))
#         disc = pt.synthetic_images.disk((5,10))
#         res = pt.zconv2(ramp, disc, 3)
#         self.assertTrue(pt.compareRecon(mres, res))
#     def test5(self):
#         matPyr = scipy.io.loadmat(op.join(matfiles_path, 'zconv2_5.mat'))
#         mres = matPyr['res']
#         ramp = pt.synthetic_images.ramp((20,10))
#         disc = pt.synthetic_images.disk((10,5))
#         res = pt.zconv2(ramp, disc, 3)
#         self.assertTrue(pt.compareRecon(mres, res))

class TestImshow(unittest.TestCase):

    def test_imshow0(self):
        im = np.random.rand(3, 10, 10)
        fig = pt.imshow([i for i in im])
        assert len(fig.axes) == 3, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow0_fail(self):
        im = np.random.rand(3, 10, 10)
        with self.assertRaises(Exception):
            # because we now need to pass a list of arrays for multiple images
            fig = pt.imshow(im)

    def test_imshow1(self):
        im = np.random.rand(3, 10, 10, 4)
        fig = pt.imshow([i for i in im])
        assert len(fig.axes) == 3, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow1_fails0(self):
        im = np.random.rand(3, 10, 10, 4)
        with self.assertRaises(Exception):
            # because we now need to pass a list of arrays for multiple images
            fig = pt.imshow(im)

    def test_imshow2(self):
        im = np.random.rand(3, 10, 10, 4)
        fig = pt.imshow(im[0])
        assert len(fig.axes) == 1, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow3(self):
        im = np.random.rand(3, 10, 10, 4)
        fig = pt.imshow([im[0, 0], im[0, 1]])
        assert len(fig.axes) == 2, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow4(self):
        im = np.random.rand(10, 10, 4)
        im2 = np.random.rand(5, 5, 4)
        fig = pt.imshow([im, im2])
        assert len(fig.axes) == 2, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow5(self):
        im = np.random.rand(2, 10, 10, 4)
        fig = pt.imshow([im[0], im[1]])
        assert len(fig.axes) == 2, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow6(self):
        im = np.random.rand(10, 10, 4)
        fig = pt.imshow([i for i in im])
        assert len(fig.axes) == 10, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow7(self):
        im = np.random.rand(10, 10)
        im2 = np.random.rand(10, 10, 4)
        fig = pt.imshow([im, im2])
        assert len(fig.axes) == 2, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow8(self):
        im = np.random.rand(2, 10, 10, 4)
        fig = pt.imshow([im[0,...,0], im[0,...,1]])
        assert len(fig.axes) == 2, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow9(self):
        im = np.random.rand(2, 10, 10, 4)
        fig = pt.imshow([im[0,...,0], im[0,...,1]])
        assert len(fig.axes) == 2, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow10(self):
        im = np.random.rand(10, 10)
        im2 = np.random.rand(5, 5, 4)
        fig = pt.imshow([im, im2])
        assert len(fig.axes) == 2, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow11(self):
        im = np.random.rand(5, 5)
        im2 = np.random.rand(10, 10, 4)
        fig = pt.imshow([im, im2])
        assert len(fig.axes) == 2, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow12(self):
        im = np.random.rand(3, 10, 10, 4)
        with self.assertRaises(Exception):
            # no longer support 4d arrays
            fig = pt.imshow(im)

    def test_imshow13(self):
        im = np.random.rand(10, 10, 3)
        im2 = np.random.rand(10, 10, 4)
        fig = pt.imshow([im, im2])
        assert len(fig.axes) == 2, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow14(self):
        im = np.random.rand(5, 5, 3)
        im2 = np.random.rand(10, 10, 4)
        fig = pt.imshow([im, im2])
        assert len(fig.axes) == 2, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_imshow15(self):
        # don't support 1d arrays
        im = np.random.rand(10)
        with self.assertRaises(Exception):
            fig = pt.imshow(im)

    def test_imshow16(self):
        # must be an array or list of them
        im = 10
        with self.assertRaises(TypeError):
            fig = pt.imshow(im)

    def test_imshow17(self):
        im = np.random.rand(256, 256) + 1j * np.random.rand(256, 256)
        fig = pt.imshow(im)
        assert len(fig.axes) == 2, "Created wrong number of axes!"

    def test_imshow18(self):
        im = np.random.rand(2, 256, 256, 3) + 1j * np.random.rand(2, 256, 256, 3)
        for c in ['rectangular', 'polar', 'logpolar']:
            fig = pt.imshow([i for i in im], plot_complex=c)
            assert len(fig.axes) == 4, "Created wrong number of axes!"

    def test_imshow19(self):
        im = np.random.randn(2, 32, 32)
        for rng in ['auto', 'indep']:
            for i in range(4):
                vrange = rng + str(i)
                pt.imshow(list(im), vrange=vrange)

class TestAnimshow(unittest.TestCase):

    def test_animshow0(self):
        vid = np.random.randn(3, 10, 10, 10)
        fig = pt.animshow([i for i in vid], as_html5=False)._fig
        assert len(fig.axes) == 3, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_animshow1(self):
        vid = np.random.rand(3, 10, 10, 10, 4)
        fig = pt.animshow(vid[0], as_html5=False)._fig
        assert len(fig.axes) == 1, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_animshow2(self):
        vid1 = np.random.rand(10, 10, 10)
        vid2 = np.random.rand(10, 5, 5)
        fig = pt.animshow([vid1, vid2], as_html5=False)._fig
        assert len(fig.axes) == 2, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_animshow3(self):
        vid1 = np.random.rand(3, 10, 10, 10)
        vid2 = np.random.rand(2, 10, 5, 5)
        fig = pt.animshow([v for v in vid1] + [v for v in vid2], as_html5=False)._fig
        assert len(fig.axes) == 5, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_animshow4(self):
        vid1 = np.random.rand(10, 10, 10, 4)
        vid2 = np.random.rand(10, 5, 5, 4)
        fig = pt.animshow([vid1, vid2], as_html5=False)._fig
        assert len(fig.axes) == 2, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_animshow5(self):
        vid = np.random.rand(6, 10, 10, 10, 4)
        fig = pt.animshow([v for v in vid], col_wrap=3, as_html5=False)._fig
        assert len(fig.axes) == 6, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_animshow6(self):
        vid = np.random.randn(2, 10, 10, 10) +\
              1j * np.random.randn(2, 10, 10, 10)
        fig = pt.animshow([v for v in vid], plot_complex='polar',
                          col_wrap=2, as_html5=False)._fig
        assert len(fig.axes) == 4, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_animshow7(self):
        vid = np.random.randn(2, 32, 32, 32, 4) +\
              1j * np.random.randn(2, 32, 32, 32, 4)
        fig = pt.animshow([v for v in vid], framerate=24,
                          plot_complex='logpolar', col_wrap=2, zoom=8,
                          title=[None, 'hello'], as_html5=False)._fig
        assert len(fig.axes) == 4, "Created wrong number of axes! Probably plotting color as grayscale or vice versa"

    def test_animshow_fail_n_frames(self):
        # raises exception because two videos have different numbers of frames
        vid1 = np.random.rand(10, 10, 10)
        vid2 = np.random.rand(5, 10, 10)
        with self.assertRaises(Exception):
            fig = pt.animshow([vid1, vid2], as_html5=False)._fig


class TestPyrshow(unittest.TestCase):

    def test_pyrshow_1d(self):
        signal = np.random.rand(256)
        pyr = pt.pyramids.GaussianPyramid(signal)
        pt.pyrshow(pyr.pyr_coeffs)

    def test_pyrshow_1d_weird_shape(self):
        # unlike 2d pyrshow, 1d pyrshow works with any shapes
        signal = np.random.rand(255)
        pyr = pt.pyramids.GaussianPyramid(signal)
        pt.pyrshow(pyr.pyr_coeffs)

    def test_pyrshow_2d(self):
        img = np.random.rand(256, 256)
        pyr = pt.pyramids.GaussianPyramid(img)
        pt.pyrshow(pyr.pyr_coeffs)

    def test_pyrshow_2d_shape_err(self):
        img = np.random.rand(255, 255)
        pyr = pt.pyramids.GaussianPyramid(img)
        with self.assertRaises(ValueError):
            pt.pyrshow(pyr.pyr_coeffs)

            
def main():
    unittest.main()

if __name__ == '__main__':
    if not os.path.exists(matfiles_path):
        print("matfiles required for testing not found, downloading now...")
        # Streaming, so we can iterate over the response.
        r = requests.get("https://osf.io/download/cbux8/", stream=True)

        # Total size in bytes.
        total_size = int(r.headers.get('content-length', 0))
        block_size = 1024*1024
        wrote = 0
        with open(matfiles_path + ".tar.gz", 'wb') as f:
            for data in tqdm.tqdm(r.iter_content(block_size), total=math.ceil(total_size//block_size),
                                  unit='MB', unit_scale=True):
                wrote = wrote  + len(data)
                f.write(data)
        if total_size != 0 and wrote != total_size:
            raise Exception("Error downloading test matfiles!")
        with tarfile.open(matfiles_path + ".tar.gz") as f:
            def is_within_directory(directory, target):
                
                abs_directory = os.path.abspath(directory)
                abs_target = os.path.abspath(target)
            
                prefix = os.path.commonprefix([abs_directory, abs_target])
                
                return prefix == abs_directory
            
            def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
            
                for member in tar.getmembers():
                    member_path = os.path.join(path, member.name)
                    if not is_within_directory(path, member_path):
                        raise Exception("Attempted Path Traversal in Tar File")
            
                tar.extractall(path, members, numeric_owner=numeric_owner) 
                
            
            safe_extract(f, os.path.dirname(matfiles_path))
        os.remove(matfiles_path + ".tar.gz")
    main()



import numpy as np
from .utils import matlab_histo


def entropy(vec, binsize=None):
    '''Compute the first-order sample entropy of `vec`

    Samples of `vec` are first discretized.  Optional argument `binsize` controls the
    discretization, and defaults to 256/(max(`vec`)-min(`vec`)).

    NOTE: This is a heavily biased estimate of entropy when you don't have much data.

    Arguments
    ---------
    vec : `array_like`
        the 2d or 2d array to calculate the entropy of
    binsize : `float` or None
        the size of the bins we discretize into. If None, will set to 256/(max(vec)-min(vec))

    Returns
    -------
    entropy : `float`
        estimate of entropy from the data

    '''
    [bincount, _] = matlab_histo(vec, nbins=256, binsize=binsize)

    # Collect non-zero bins:
    H = bincount[np.where(bincount > 0)]
    H = H / H.sum()

    return -(H * np.log2(H)).sum()


def range(array):
    '''compute minimum and maximum values of the input array

    `array` must be real-valued

    Arguments
    ---------
    array : `np.array`
        array to calculate the range of

    Returns
    -------
    array_range : `tuple`
        (min, max)
    '''
    if not np.isreal(array.all()):
        raise Exception('array must be real-valued')

    return (array.min(), array.max())


def var(array, array_mean=None):
    '''Sample variance of the input numpy array.

    Passing `mean` (optional) makes the calculation faster. This works equally well for real and
    complex-valued `array`

    Arguments
    ---------
    array : `np.array`
        array to calculate the variance of
    array_mean : `float` or None
        the mean of `array`. If None, will calculate it.

    Returns
    -------
    array_var : `float`
        the variance of `array`
    '''
    if array_mean is None:
        array_mean = array.mean()

    if np.isreal(array).all():
        return ((array - array_mean)**2).sum() / max(array.size - 1, 1)
    else:
        return var(array.real, array_mean.real) + 1j * var(array.imag, array_mean.imag)


def skew(array, array_mean=None, array_var=None):
    '''Sample skew (third moment divided by variance^3/2) of the input array.

    `mean` (optional) and `var` (optional) make the computation faster. This works equally well for
    real and complex-valued `array`

    Arguments
    ---------
    array : `np.array`
        array to calculate the variance of
    array_mean : `float` or None
        the mean of `array`. If None, will calculate it.
    array_var : `float` or None
        the variance of `array`. If None, will calculate it

    Returns
    -------
    array_skew : `float`
        the skew of `array`.

    '''
    if array_mean is None:
        array_mean = array.mean()
    if array_var is None:
        array_var = var(array, array_mean)

    if np.isreal(array).all():
        return ((array - array_mean)**3).mean() / np.sqrt(array_var) ** 3
    else:
        return (skew(array.real, array_mean.real, array_var.real) + 1j *
                skew(array.imag, array_mean.imag, array_var.imag))


def kurt(array, array_mean=None, array_var=None):
    '''Sample kurtosis (fourth moment divided by squared variance) of the input array.

    For reference, kurtosis of a Gaussian distribution is 3.

    `mean` (optional) and `var` (optional) make the computation faster. This works equally well for
    real and complex-valued `array`

    Arguments
    ---------
    array : `np.array`
        array to calculate the variance of
    array_mean : `float` or None
        the mean of `array`. If None, will calculate it.
    array_var : `float` or None
        the variance of `array`. If None, will calculate it

    Returns
    -------
    array_kurt : `float`
        the kurtosis of `array`.

    '''
    if array_mean is None:
        array_mean = array.mean()
    if array_var is None:
        array_var = var(array, array_mean)

    if np.isreal(array).all():
        return ((array - array_mean) ** 4).mean() / array_var ** 2
    else:
        return (kurt(array.real, array_mean.real, array_var.real) + 1j *
                kurt(array.imag, array_mean.imag, array_var.imag))


def image_compare(im_array0, im_array1):
    '''Prints and returns min, max, mean, stdev of the difference, and SNR (relative to im_array0).

    Arguments
    ---------
    im_array0 : `np.array`
        the first image to compare
    im_array1 : `np.array`
        the second image to compare

    Returns
    -------
    min_diff : `float`
        the minimum difference between `im_array0` and `im_array1`
    max_diff : `float`
        the maximum difference between `im_array0` and `im_array1`
    mean_diff : `float`
        the mean difference between `im_array0` and `im_array1`
    std_diff : `float`
        the standard deviation of the difference between `im_array0` and `im_array1`
    snr : `float`
        the signal-to-noise ratio of the difference between `im_array0` and `im_array0` (relative
        to `im_array0`)
    '''
    if not im_array0.size == im_array1.size:
        raise Exception('Input images must have the same size')

    if not np.isreal(im_array0).all() or not np.isreal(im_array1).all():
        raise Exception('Input images must be real-valued matrices')

    difference = im_array0 - im_array1
    (min_diff, max_diff) = range(difference)
    mean_diff = difference.mean()
    var_diff = var(difference, mean_diff)
    if var_diff < np.finfo(np.double).tiny:
        snr = np.inf
    else:
        snr = 10 * np.log10(var(im_array0) / var_diff)
    print('Difference statistics:')
    print('  Range: [%d, %d]' % (min_diff, max_diff))
    print('  Mean: %f,  Stdev (rmse): %f,  SNR (dB): %f' % (mean_diff, np.sqrt(var_diff), snr))
    return min_diff, max_diff, mean_diff, np.sqrt(var_diff), snr


def image_stats(im_array):
    '''Prints and returns image statistics: min, max, mean, stdev, and kurtosis.

    Arguments
    ---------
    im_array : `np.array`
        the image to summarize

    Returns
    -------
    array_min : `float`
        the minimum of `im_array`
    array_max : `float`
        the maximum of `im_array`
    array_mean : `float`
        the mean of `im_array`
    array_std : `float`
        the standard deviation of `im_array`
    array_kurt : `float`
        the kurtosis of `im_array`
    '''
    if not np.isreal(im_array).all():
        raise Exception('Input images must be real-valued matrices')

    (mini, maxi) = range(im_array)
    array_mean = im_array.mean()
    array_var = var(im_array, array_mean)
    array_kurt = kurt(im_array, array_mean, array_var)
    print('Image statistics:')
    print('  Range: [%f, %f]' % (mini, maxi))
    print('  Mean: %f,  Stdev: %f,  Kurtosis: %f' % (array_mean, np.sqrt(array_var), array_kurt))
    return mini, maxi, array_mean, np.sqrt(array_var), array_kurt



from .pyramid import Pyramid
from .filters import parse_filter
from .c.wrapper import corrDn


class GaussianPyramid(Pyramid):
    """Gaussian pyramid

    Parameters
    ----------
    image : `array_like`
        1d or 2d image upon which to construct to the pyramid.
    height : 'auto' or `int`.
        The height of the pyramid. If 'auto', will automatically determine based on the size of
        `image`.
    filter_name : {'binomN', 'haar', 'qmf8', 'qmf12', 'qmf16', 'daub2', 'daub3', 'daub4', 'qmf5',
                   'qmf9', 'qmf13'}
        name of filter to use when constructing pyramid. All scaled so L-2 norm is 1.0

        * `'binomN'` - binomial coefficient filter of order N-1
        * `'haar'` - Haar wavelet
        * `'qmf8'`, `'qmf12'`, `'qmf16'` - Symmetric Quadrature Mirror Filters [1]_
        * `'daub2'`, `'daub3'`, `'daub4'` - Daubechies wavelet [2]_
        * `'qmf5'`, `'qmf9'`, `'qmf13'`   - Symmetric Quadrature Mirror Filters [3]_, [4]_
    edge_type : {'circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend', 'dont-compute'}
        Specifies how to handle edges. Options are:

        * `'circular'` - circular convolution
        * `'reflect1'` - reflect about the edge pixels
        * `'reflect2'` - reflect, doubling the edge pixels
        * `'repeat'` - repeat the edge pixels
        * `'zero'` - assume values of zero outside image boundary
        * `'extend'` - reflect and invert
        * `'dont-compute'` - zero output when filter overhangs imput boundaries.

    Attributes
    ----------
    image : `array_like`
        The input image used to construct the pyramid.
    image_size : `tuple`
        The size of the input image.
    pyr_type : `str` or `None`
        Human-readable string specifying the type of pyramid. For base class, is None.
    edge_type : `str`
        Specifies how edges were handled.
    pyr_coeffs : `dict`
        Dictionary containing the coefficients of the pyramid. Keys are `(level, band)` tuples and
        values are 1d or 2d numpy arrays (same number of dimensions as the input image)
    pyr_size : `dict`
        Dictionary containing the sizes of the pyramid coefficients. Keys are `(level, band)`
        tuples and values are tuples.
    is_complex : `bool`
        Whether the coefficients are complex- or real-valued. Only `SteerablePyramidFreq` can have
        a value of True, all others must be False.

    References
    ----------
    .. [1] J D Johnston, "A filter family designed for use in quadrature mirror filter banks",
       Proc. ICASSP, pp 291-294, 1980.
    .. [2] I Daubechies, "Orthonormal bases of compactly supported wavelets", Commun. Pure Appl.
       Math, vol. 42, pp 909-996, 1988.
    .. [3] E P Simoncelli,  "Orthogonal sub-band image transforms", PhD Thesis, MIT Dept. of Elec.
       Eng. and Comp. Sci. May 1988. Also available as: MIT Media Laboratory Vision and Modeling
       Technical Report #100.
    .. [4] E P Simoncelli and E H Adelson, "Subband image coding", Subband Transforms, chapter 4,
       ed. John W Woods, Kluwer Academic Publishers,  Norwell, MA, 1990, pp 143--192.

    """

    def __init__(self, image, height='auto', filter_name='binom5', edge_type='reflect1', **kwargs):
        super().__init__(image=image, edge_type=edge_type)
        if self.pyr_type is None:
            self.pyr_type = 'Gaussian'
        self.num_orientations = 1

        self.filters = {'downsample_filter': parse_filter(filter_name, normalize=False)}
        upsamp_filt = kwargs.pop('upsample_filter_name', None)
        if upsamp_filt is not None:
            if self.pyr_type != 'Laplacian':
                raise Exception("upsample_filter should only be set for Laplacian pyramid!")
            self.filters['upsample_filter'] = parse_filter(upsamp_filt, normalize=False)
        self._set_num_scales('downsample_filter', height, 1)

        self._build_pyr()

    def _build_next(self, image):
        """build the next level of the pyramid

        This should not be called directly by users, it's a helper function for constructing the
        pyramid

        """
        if image.shape[0] == 1:
            res = corrDn(image=image, filt=self.filters['downsample_filter'].T, edge_type=self.edge_type, step=(1, 2))
        elif image.shape[1] == 1:
            res = corrDn(image=image, filt=self.filters['downsample_filter'], edge_type=self.edge_type, step=(2, 1))
        else:
            tmp = corrDn(image=image, filt=self.filters['downsample_filter'].T, edge_type=self.edge_type, step=(1, 2))
            res = corrDn(image=tmp, filt=self.filters['downsample_filter'], edge_type=self.edge_type, step=(2, 1))
        return res

    def _build_pyr(self):
        """build the pyramid

        This should not be called directly by users, it's a helper function for constructing the
        pyramid

        we do this in a separate method for a bit of class wizardry: by over-writing this method in
        the LaplacianPyramid class, which inherits the GaussianPyramid class, we can still
        correctly construct the LaplacianPyramid with a single call to the GaussianPyramid
        constructor
        """
        im = self.image
        self.pyr_coeffs[(0, 0)] = self.image.copy()
        self.pyr_size[(0, 0)] = self.image_size
        for lev in range(1, self.num_scales):
            im = self._build_next(im)
            self.pyr_coeffs[(lev, 0)] = im.copy()
            self.pyr_size[(lev, 0)] = im.shape

    def recon_pyr(self, *args):
        """Reconstruct the pyramid -- NOT NECESSARY FOR GAUSSIANS
        """
        raise Exception('Not necessary for Gaussian Pyramids')



import warnings
import numpy as np
from scipy.special import factorial
from .pyramid import SteerablePyramidBase
from .c.wrapper import pointOp
from ..tools.utils import rcosFn


class SteerablePyramidFreq(SteerablePyramidBase):
    """Steerable frequency pyramid.

    Construct a steerable pyramid on matrix IM, in the Fourier domain.
    This is similar to Spyr, except that:

        + Reconstruction is exact (within floating point errors)
        + It can produce any number of orientation bands.
        - Typically slower, especially for non-power-of-two sizes.
        - Boundary-handling is circular.

    The squared radial functions tile the Fourier plane with a raised-cosine
    falloff. Angular functions are cos(theta- k*pi/order+1)^(order).

    Note that reconstruction will not be exact if the image has an odd shape (due to
    boundary-handling issues) or if the pyramid is complex with order=0.

    Notes
    -----
    Transform described in [1]_, filter kernel design described in [2]_.

    Parameters
    ----------
    image : `array_like`
        2d image upon which to construct to the pyramid.
    height : 'auto' or `int`.
        The height of the pyramid. If 'auto', will automatically determine based on the size of
        `image`. If an int, must be non-negative. When height=0, only returns the residuals.
    order : `int`.
        The Gaussian derivative order used for the steerable filters. Default value is 3.
        Note that to achieve steerability the minimum number of orientation is `order` + 1,
        and is used here. To get more orientations at the same order, use the method `steer_coeffs`
    twidth : `int`
        The width of the transition region of the radial lowpass function, in octaves
    is_complex : `bool`
        Whether the pyramid coefficients should be complex or not. If True, the real and imaginary
        parts correspond to a pair of odd and even symmetric filters. If False, the coefficients
        only include the real part / odd symmetric filter.

    Attributes
    ----------
    image : `array_like`
        The input image used to construct the pyramid.
    image_size : `tuple`
        The size of the input image.
    pyr_type : `str` or `None`
        Human-readable string specifying the type of pyramid. For base class, is None.
    pyr_coeffs : `dict`
        Dictionary containing the coefficients of the pyramid. Keys are `(level, band)` tuples and
        values are 1d or 2d numpy arrays (same number of dimensions as the input image),
        running from fine to coarse.
    pyr_size : `dict`
        Dictionary containing the sizes of the pyramid coefficients. Keys are `(level, band)`
        tuples and values are tuples.
    is_complex : `bool`
        Whether the coefficients are complex- or real-valued.

    References
    ----------
    .. [1] E P Simoncelli and W T Freeman, "The Steerable Pyramid: A Flexible Architecture for
       Multi-Scale Derivative Computation," Second Int'l Conf on Image Processing, Washington, DC,
       Oct 1995.
    .. [2] A Karasaridis and E P Simoncelli, "A Filter Design Technique for Steerable Pyramid
       Image Transforms", ICASSP, Atlanta, GA, May 1996.

    """
    def __init__(self, image, height='auto', order=3, twidth=1, is_complex=False):
        # in the Fourier domain, there's only one choice for how do edge-handling: circular. to
        # emphasize that thisisn'ta choice, we use None here.
        super().__init__(image=image, edge_type=None)

        self.pyr_type = 'SteerableFrequency'
        self.is_complex = is_complex
        # SteerablePyramidFreq doesn't have filters, they're constructed in the frequency space
        self.filters = {}
        self.order = int(order)

        if (image.shape[0] % 2 != 0) or (image.shape[1] % 2 != 0):
            warnings.warn("Reconstruction will not be perfect with odd-sized images")

        if self.order == 0 and self.is_complex:
            raise ValueError(
                "Complex pyramid cannot have order=0! See "
                "https://github.com/plenoptic-org/plenoptic/issues/326 "
                "for an explanation."
            )

        # we can't use the base class's _set_num_scales method because the max height is calculated
        # slightly differently
        max_ht = np.floor(np.log2(min(self.image.shape))) - 2
        if height == 'auto' or height is None:
            self.num_scales = int(max_ht)
        elif height > max_ht:
            raise ValueError("Cannot build pyramid higher than %d levels." % (max_ht))
        elif height < 0:
            raise ValueError("Height must be a non-negative int.")
        else:
            self.num_scales = int(height)

        if self.order > 15 or self.order < 0:
            raise ValueError("order must be an integer in the range [0,15].")

        self.num_orientations = int(order + 1)

        if twidth <= 0:
            raise ValueError("twidth must be positive.")
        twidth = int(twidth)

        dims = np.asarray(self.image.shape)
        ctr = np.ceil((np.asarray(dims)+0.5)/2).astype(int)

        (xramp, yramp) = np.meshgrid(np.linspace(-1, 1, dims[1]+1)[:-1],
                                     np.linspace(-1, 1, dims[0]+1)[:-1])

        angle = np.arctan2(yramp, xramp)
        log_rad = np.sqrt(xramp**2 + yramp**2)
        log_rad[ctr[0]-1, ctr[1]-1] = log_rad[ctr[0]-1, ctr[1]-2]
        log_rad = np.log2(log_rad)

        # Radial transition function (a raised cosine in log-frequency):
        (Xrcos, Yrcos) = rcosFn(twidth, (-twidth/2.0), np.asarray([0, 1]))
        Yrcos = np.sqrt(Yrcos)

        YIrcos = np.sqrt(1.0 - Yrcos**2)
        lo0mask = pointOp(log_rad, YIrcos, Xrcos[0], Xrcos[1]-Xrcos[0])
        self._lo0mask = lo0mask

        imdft = np.fft.fftshift(np.fft.fft2(self.image))

        hi0mask = pointOp(log_rad, Yrcos, Xrcos[0], Xrcos[1]-Xrcos[0])
        self._hi0mask = hi0mask

        hi0dft = imdft * hi0mask.reshape(imdft.shape[0], imdft.shape[1])
        hi0 = np.fft.ifft2(np.fft.ifftshift(hi0dft))

        self.pyr_coeffs['residual_highpass'] = np.real(hi0)
        self.pyr_size['residual_highpass'] = hi0.shape

        lo0mask = lo0mask.reshape(imdft.shape[0], imdft.shape[1])
        lodft = imdft * lo0mask

        self._anglemasks = []
        self._himasks = []
        self._lomasks = []

        for i in range(self.num_scales):
            Xrcos -= np.log2(2)

            lutsize = 1024
            Xcosn = np.pi * np.arange(-(2*lutsize+1), (lutsize+2)) / lutsize

            const = (2**(2*self.order))*(factorial(self.order, exact=True)**2)/ float(self.num_orientations*factorial(2*self.order, exact=True))

            if self.is_complex:
                # TODO clean that up and give comments
                alfa = ((np.pi+Xcosn) % (2.0*np.pi)) - np.pi
                Ycosn = (2.0 * np.sqrt(const) * (np.cos(Xcosn) ** self.order) *
                         (np.abs(alfa) < np.pi/2.0).astype(int))
            else:
                Ycosn = np.sqrt(const) * (np.cos(Xcosn))**self.order

            log_rad_test = np.reshape(log_rad, (1, log_rad.shape[0] * log_rad.shape[1]))
            himask = pointOp(log_rad_test, Yrcos, Xrcos[0], Xrcos[1]-Xrcos[0])
            himask = himask.reshape((lodft.shape[0], lodft.shape[1]))
            self._himasks.append(himask)

            anglemasks = []
            for b in range(self.num_orientations):
                angle_tmp = np.reshape(angle, (1, angle.shape[0] * angle.shape[1]))
                anglemask = pointOp(angle_tmp, Ycosn, Xcosn[0]+np.pi*b/self.num_orientations,
                                    Xcosn[1]-Xcosn[0])

                anglemask = anglemask.reshape(lodft.shape[0], lodft.shape[1])
                anglemasks.append(anglemask)
                # that (-1j)**order term in the beginning will be 1, -j, -1, j for order 0, 1, 2,
                # 3, and will then loop again
                banddft = (-1j) ** self.order * lodft * anglemask * himask
                band = np.fft.ifft2(np.fft.ifftshift(banddft))
                if not self.is_complex:
                    self.pyr_coeffs[(i, b)] = np.real(band.copy())
                else:
                    self.pyr_coeffs[(i, b)] = band.copy()
                self.pyr_size[(i, b)] = band.shape

            self._anglemasks.append(anglemasks)
            dims = np.asarray(lodft.shape)
            ctr = np.ceil((dims+0.5)/2).astype(int)
            lodims = np.ceil((dims-0.5)/2).astype(int)
            loctr = np.ceil((lodims+0.5)/2).astype(int)
            lostart = ctr - loctr
            loend = lostart + lodims

            log_rad = log_rad[lostart[0]:loend[0], lostart[1]:loend[1]]
            angle = angle[lostart[0]:loend[0], lostart[1]:loend[1]]
            lodft = lodft[lostart[0]:loend[0], lostart[1]:loend[1]]
            YIrcos = np.abs(np.sqrt(1.0 - Yrcos**2))
            log_rad_tmp = np.reshape(log_rad, (1, log_rad.shape[0] * log_rad.shape[1]))
            lomask = pointOp(log_rad_tmp, YIrcos, Xrcos[0], Xrcos[1]-Xrcos[0])
            lomask = lomask.reshape(lodft.shape[0], lodft.shape[1])
            self._lomasks.append(lomask)

            lodft = lodft * lomask

        lodft = np.fft.ifft2(np.fft.ifftshift(lodft))
        self.pyr_coeffs['residual_lowpass'] = np.real(np.asarray(lodft).copy())
        self.pyr_size['residual_lowpass'] = lodft.shape

    def recon_pyr(self, levels='all', bands='all', twidth=1):
        """Reconstruct the image, optionally using subset of pyramid coefficients.

        Parameters
        ----------
        levels : `list`, `int`,  or {`'all'`, `'residual_highpass'`}
            If `list` should contain some subset of integers from `0` to `self.num_scales-1`
            (inclusive) and `'residual_lowpass'`. If `'all'`, returned value will contain all
            valid levels. Otherwise, must be one of the valid levels.
        bands : `list`, `int`, or `'all'`.
            If list, should contain some subset of integers from `0` to `self.num_orientations-1`.
            If `'all'`, returned value will contain all valid orientations. Otherwise, must be one
            of the valid orientations.
        twidth : `int`
            The width of the transition region of the radial lowpass function, in octaves

        Returns
        -------
        recon : `np.array`
            The reconstructed image.

        """
        if twidth <= 0:
            raise ValueError("twidth must be positive.")

        recon_keys = self._recon_keys(levels, bands)

        # make list of dims and bounds
        bound_list = []
        dim_list = []
        # we go through pyr_sizes from smallest to largest
        for dims in sorted(self.pyr_size.values()):
            if dims in dim_list:
                continue
            dim_list.append(dims)
            dims = np.asarray(dims)
            ctr = np.ceil((dims+0.5)/2).astype(int)
            lodims = np.ceil((dims-0.5)/2).astype(int)
            loctr = np.ceil((lodims+0.5)/2).astype(int)
            lostart = ctr - loctr
            loend = lostart + lodims
            bounds = (lostart[0], lostart[1], loend[0], loend[1])
            bound_list.append(bounds)
        bound_list.append((0, 0, dim_list[-1][0], dim_list[-1][1]))
        dim_list.append((dim_list[-1][0], dim_list[-1][1]))

        # matlab code starts here
        dims = np.asarray(self.pyr_size['residual_highpass'])
        ctr = np.ceil((dims+0.5)/2.0).astype(int)

        (xramp, yramp) = np.meshgrid((np.arange(1, dims[1]+1)-ctr[1]) / (dims[1]/2.),
                                     (np.arange(1, dims[0]+1)-ctr[0]) / (dims[0]/2.))
        angle = np.arctan2(yramp, xramp)
        log_rad = np.sqrt(xramp**2 + yramp**2)
        log_rad[ctr[0]-1, ctr[1]-1] = log_rad[ctr[0]-1, ctr[1]-2]
        log_rad = np.log2(log_rad)

        # Radial transition function (a raised cosine in log-frequency):
        (Xrcos, Yrcos) = rcosFn(twidth, (-twidth/2.0), np.asarray([0, 1]))
        Yrcos = np.sqrt(Yrcos)
        YIrcos = np.sqrt(1.0 - Yrcos**2)

        # from reconSFpyrLevs
        lutsize = 1024

        Xcosn = np.pi * np.arange(-(2*lutsize+1), (lutsize+2)) / lutsize

        const = (2**(2*self.order))*(factorial(self.order, exact=True)**2) / float(self.num_orientations*factorial(2*self.order, exact=True))
        Ycosn = np.sqrt(const) * (np.cos(Xcosn))**self.order

        # lowest band
        # initialize reconstruction
        if 'residual_lowpass' in recon_keys:
            nresdft = np.fft.fftshift(np.fft.fft2(self.pyr_coeffs['residual_lowpass']))
        else:
            nresdft = np.zeros_like(self.pyr_coeffs['residual_lowpass'])
        resdft = np.zeros(dim_list[1]) + 0j

        bounds = (0, 0, 0, 0)
        for idx in range(len(bound_list)-2, 0, -1):
            diff = (bound_list[idx][2]-bound_list[idx][0],
                    bound_list[idx][3]-bound_list[idx][1])
            bounds = (bounds[0]+bound_list[idx][0], bounds[1]+bound_list[idx][1],
                      bounds[0]+bound_list[idx][0] + diff[0],
                      bounds[1]+bound_list[idx][1] + diff[1])
            Xrcos -= np.log2(2.0)
        nlog_rad = log_rad[bounds[0]:bounds[2], bounds[1]:bounds[3]]

        nlog_rad_tmp = np.reshape(nlog_rad, (1, nlog_rad.shape[0]*nlog_rad.shape[1]))
        lomask = pointOp(nlog_rad_tmp, YIrcos, Xrcos[0], Xrcos[1]-Xrcos[0])
        lomask = lomask.reshape(nresdft.shape[0], nresdft.shape[1])
        lomask = lomask + 0j
        resdft[bound_list[1][0]:bound_list[1][2],
               bound_list[1][1]:bound_list[1][3]] = nresdft * lomask

        # middle bands
        for idx in range(1, len(bound_list)-1):
            bounds1 = (0, 0, 0, 0)
            bounds2 = (0, 0, 0, 0)
            for boundIdx in range(len(bound_list) - 1, idx - 1, -1):
                diff = (bound_list[boundIdx][2]-bound_list[boundIdx][0],
                        bound_list[boundIdx][3]-bound_list[boundIdx][1])
                bound2tmp = bounds2
                bounds2 = (bounds2[0]+bound_list[boundIdx][0],
                           bounds2[1]+bound_list[boundIdx][1],
                           bounds2[0]+bound_list[boundIdx][0] + diff[0],
                           bounds2[1]+bound_list[boundIdx][1] + diff[1])
                bounds1 = bound2tmp
            nlog_rad1 = log_rad[bounds1[0]:bounds1[2], bounds1[1]:bounds1[3]]
            nlog_rad2 = log_rad[bounds2[0]:bounds2[2], bounds2[1]:bounds2[3]]
            dims = dim_list[idx]
            nangle = angle[bounds1[0]:bounds1[2], bounds1[1]:bounds1[3]]
            YIrcos = np.abs(np.sqrt(1.0 - Yrcos**2))
            if idx > 1:
                Xrcos += np.log2(2.0)
                nlog_rad2_tmp = np.reshape(nlog_rad2, (1, nlog_rad2.shape[0]*nlog_rad2.shape[1]))
                lomask = pointOp(nlog_rad2_tmp, YIrcos, Xrcos[0],
                                 Xrcos[1]-Xrcos[0])
                lomask = lomask.reshape(bounds2[2]-bounds2[0],
                                        bounds2[3]-bounds2[1])
                lomask = lomask + 0j
                nresdft = np.zeros(dim_list[idx]) + 0j
                nresdft[bound_list[idx][0]:bound_list[idx][2],
                        bound_list[idx][1]:bound_list[idx][3]] = resdft * lomask
                resdft = nresdft.copy()

            # reconSFpyrLevs
            if idx != 0 and idx != len(bound_list)-1:
                for b in range(self.num_orientations):
                    nlog_rad1_tmp = np.reshape(nlog_rad1,
                                               (1, nlog_rad1.shape[0]*nlog_rad1.shape[1]))
                    himask = pointOp(nlog_rad1_tmp, Yrcos, Xrcos[0], Xrcos[1]-Xrcos[0])

                    himask = himask.reshape(nlog_rad1.shape)
                    nangle_tmp = np.reshape(nangle, (1, nangle.shape[0]*nangle.shape[1]))
                    anglemask = pointOp(nangle_tmp, Ycosn,
                                        Xcosn[0]+np.pi*b/self.num_orientations,
                                        Xcosn[1]-Xcosn[0])

                    anglemask = anglemask.reshape(nangle.shape)
                    # either the coefficients will already be real-valued (if
                    # self.is_complex=False) or complex (if self.is_complex=True). in the
                    # former case, this np.real() does nothing. in the latter, we want to only
                    # reconstruct with the real portion
                    curLev = self.num_scales-1 - (idx-1)
                    band = np.real(self.pyr_coeffs[(curLev, b)])
                    if (curLev, b) in recon_keys:
                        banddft = np.fft.fftshift(np.fft.fft2(band))
                    else:
                        banddft = np.zeros(band.shape)
                    resdft += ((np.power(-1+0j, 0.5))**(self.num_orientations-1) *
                               banddft * anglemask * himask)

        # apply lo0mask
        Xrcos += np.log2(2.0)
        lo0mask = pointOp(log_rad, YIrcos, Xrcos[0], Xrcos[1]-Xrcos[0])

        lo0mask = lo0mask.reshape(dims[0], dims[1])
        resdft = resdft * lo0mask

        # residual highpass subband
        hi0mask = pointOp(log_rad, Yrcos, Xrcos[0], Xrcos[1]-Xrcos[0])

        hi0mask = hi0mask.reshape(resdft.shape[0], resdft.shape[1])
        if 'residual_highpass' in recon_keys:
            hidft = np.fft.fftshift(np.fft.fft2(self.pyr_coeffs['residual_highpass']))
        else:
            hidft = np.zeros_like(self.pyr_coeffs['residual_highpass'])
        resdft += hidft * hi0mask

        outresdft = np.real(np.fft.ifft2(np.fft.ifftshift(resdft)))

        return outresdft



import numpy as np
from .pyramid import Pyramid
from .filters import parse_filter
from .c.wrapper import corrDn, upConv


class WaveletPyramid(Pyramid):
    """Multiscale wavelet pyramid

    Parameters
    ----------
    image : `array_like`
        1d or 2d image upon which to construct to the pyramid.
    height : 'auto' or `int`.
        The height of the pyramid. If 'auto', will automatically determine based on the size of
        `image`.
    filter_name : {'binomN', 'haar', 'qmf8', 'qmf12', 'qmf16', 'daub2', 'daub3', 'daub4', 'qmf5',
                   'qmf9', 'qmf13'}
        name of filter to use when constructing pyramid. All scaled so L-2 norm is 1.0

        * `'binomN'` - binomial coefficient filter of order N-1
        * `'haar'` - Haar wavelet
        * `'qmf8'`, `'qmf12'`, `'qmf16'` - Symmetric Quadrature Mirror Filters [1]_
        * `'daub2'`, `'daub3'`, `'daub4'` - Daubechies wavelet [2]_
        * `'qmf5'`, `'qmf9'`, `'qmf13'`   - Symmetric Quadrature Mirror Filters [3]_, [4]_
    edge_type : {'circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend', 'dont-compute'}
        Specifies how to handle edges. Options are:

        * `'circular'` - circular convolution
        * `'reflect1'` - reflect about the edge pixels
        * `'reflect2'` - reflect, doubling the edge pixels
        * `'repeat'` - repeat the edge pixels
        * `'zero'` - assume values of zero outside image boundary
        * `'extend'` - reflect and invert
        * `'dont-compute'` - zero output when filter overhangs imput boundaries.

    Attributes
    ----------
    image : `array_like`
        The input image used to construct the pyramid.
    image_size : `tuple`
        The size of the input image.
    pyr_type : `str` or `None`
        Human-readable string specifying the type of pyramid. For base class, is None.
    edge_type : `str`
        Specifies how edges were handled.
    pyr_coeffs : `dict`
        Dictionary containing the coefficients of the pyramid. Keys are `(level, band)` tuples and
        values are 1d or 2d numpy arrays (same number of dimensions as the input image)
    pyr_size : `dict`
        Dictionary containing the sizes of the pyramid coefficients. Keys are `(level, band)`
        tuples and values are tuples.
    is_complex : `bool`
        Whether the coefficients are complex- or real-valued. Only `SteerablePyramidFreq` can have
        a value of True, all others must be False.

    References
    ----------
    .. [1] J D Johnston, "A filter family designed for use in quadrature mirror filter banks",
       Proc. ICASSP, pp 291-294, 1980.
    .. [2] I Daubechies, "Orthonormal bases of compactly supported wavelets", Commun. Pure Appl.
       Math, vol. 42, pp 909-996, 1988.
    .. [3] E P Simoncelli,  "Orthogonal sub-band image transforms", PhD Thesis, MIT Dept. of Elec.
       Eng. and Comp. Sci. May 1988. Also available as: MIT Media Laboratory Vision and Modeling
       Technical Report #100.
    .. [4] E P Simoncelli and E H Adelson, "Subband image coding", Subband Transforms, chapter 4,
       ed. John W Woods, Kluwer Academic Publishers,  Norwell, MA, 1990, pp 143--192.
    """

    def __init__(self, image, height='auto', filter_name='qmf9', edge_type='reflect1'):
        super().__init__(image=image, edge_type=edge_type)
        self.pyr_type = 'Wavelet'

        self.filters = {}
        self.filters['lo_filter'] = parse_filter(filter_name, normalize=False)
        self.filters["hi_filter"] = WaveletPyramid._modulate_flip(self.filters['lo_filter'])
        assert self.filters['lo_filter'].shape == self.filters['hi_filter'].shape

        # Stagger sampling if filter is odd-length
        self.stagger = (self.filters['lo_filter'].size + 1) % 2

        self._set_num_scales('lo_filter', height)

        # compute the number of channels per level
        if min(self.image.shape) == 1:
            self.num_orientations = 1
        else:
            self.num_orientations = 3

        self._build_pyr()

    def _modulate_flip(lo_filter):
        '''construct QMF/Wavelet highpass filter from lowpass filter

        modulate by (-1)^n, reverse order (and shift by one, which is handled by the convolution
        routines).  This is an extension of the original definition of QMF's (e.g., see
        Simoncelli90).

        Parameters
        ----------
        lo_filter : `array_like`
            one-dimensional array (or effectively 1d array) containing the lowpass filter to
            convert into the highpass filter.

        Returns
        -------
        hi_filter : `np.array`
            The highpass filter constructed from the lowpass filter, same shape as the lowpass
            filter.
        '''
        # check lo_filter is effectively 1D
        lo_filter_shape = lo_filter.shape
        assert lo_filter.size == max(lo_filter_shape)
        lo_filter = lo_filter.flatten()
        ind = np.arange(lo_filter.size, 0, -1) - (lo_filter.size + 1) // 2
        hi_filter = lo_filter[::-1] * (-1.0) ** ind

        return hi_filter.reshape(lo_filter_shape)

    def _build_next(self, image):
        """Build the next level fo the Wavelet pyramid

        Should not be called by users directly, this is a helper function to construct the pyramid.

        Parameters
        ----------
        image : `array_like`
            image to use to construct next level.

        Returns
        -------
        lolo : `array_like`
            This is the result of applying the lowpass filter once if `image` is 1d, twice if it's
            2d. It's downsampled by a factor of two from the original `image`.
        hi_tuple : `tuple`
            If `image` is 1d, this just contains `hihi`, the result of applying the highpass filter
            . If `image` is 2d, it is `(lohi, hilo, hihi)`, the result of applying the lowpass then
            the highpass, the highpass then the lowpass, and the highpass twice. All will be
            downsampled by a factor of two from the original `image`.
        """
        if image.shape[1] == 1:
            lolo = corrDn(image=image, filt=self.filters['lo_filter'], edge_type=self.edge_type, step=(2, 1), start=(self.stagger, 0))
            hihi = corrDn(image=image, filt=self.filters['hi_filter'], edge_type=self.edge_type, step=(2, 1), start=(1, 0))
            return lolo, (hihi, )
        elif image.shape[0] == 1:
            lolo = corrDn(image=image, filt=self.filters['lo_filter'].T, edge_type=self.edge_type, step=(1, 2), start=(0, self.stagger))
            hihi = corrDn(image=image, filt=self.filters['hi_filter'].T, edge_type=self.edge_type, step=(1, 2), start=(0, 1))
            return lolo, (hihi, )
        else:
            lo = corrDn(image=image, filt=self.filters['lo_filter'], edge_type=self.edge_type, step=(2, 1), start=(self.stagger, 0))
            hi = corrDn(image=image, filt=self.filters['hi_filter'], edge_type=self.edge_type, step=(2, 1), start=(1, 0))
            lolo = corrDn(image=lo, filt=self.filters['lo_filter'].T, edge_type=self.edge_type, step=(1, 2), start=(0, self.stagger))
            lohi = corrDn(image=hi, filt=self.filters['lo_filter'].T, edge_type=self.edge_type, step=(1, 2), start=(0, self.stagger))
            hilo = corrDn(image=lo, filt=self.filters['hi_filter'].T, edge_type=self.edge_type, step=(1, 2), start=(0, 1))
            hihi = corrDn(image=hi, filt=self.filters['hi_filter'].T, edge_type=self.edge_type, step=(1, 2), start=(0, 1))
            return lolo, (lohi, hilo, hihi)

    def _build_pyr(self):
        im = self.image
        for lev in range(self.num_scales):
            im, higher_bands = self._build_next(im)
            for j, band in enumerate(higher_bands):
                self.pyr_coeffs[(lev, j)] = band
                self.pyr_size[(lev, j)] = band.shape
        self.pyr_coeffs['residual_lowpass'] = im
        self.pyr_size['residual_lowpass'] = im.shape


    def _recon_prev(self, image, lev, recon_keys, output_size, lo_filter, hi_filter, edge_type,
                    stagger):
        """Reconstruct the previous level of the pyramid.

        Should not be called by users directly, this is a helper function for reconstructing the
        input image using pyramid coefficients.

        """
        if self.num_orientations == 1:
            if output_size[0] == 1:
                recon = upConv(image=image, filt=lo_filter.T, edge_type=edge_type, step=(1, 2), start=(0, stagger), stop=output_size)
                if (lev, 0) in recon_keys:
                    recon += upConv(image=self.pyr_coeffs[(lev, 0)], filt=hi_filter.T, edge_type=edge_type, step=(1, 2), start=(0, 1), stop=output_size)
            elif output_size[1] == 1:
                recon = upConv(image=image, filt=lo_filter, edge_type=edge_type, step=(2, 1), start=(stagger, 0), stop=output_size)
                if (lev, 0) in recon_keys:
                    recon += upConv(image=self.pyr_coeffs[(lev, 0)], filt=hi_filter, edge_type=edge_type, step=(2, 1), start=(1, 0), stop=output_size)
        else:
            lo_size = ([self.pyr_size[(lev, 1)][0], output_size[1]])
            hi_size = ([self.pyr_size[(lev, 0)][0], output_size[1]])

            tmp_recon = upConv(image=image, filt=lo_filter.T, edge_type=edge_type, step=(1, 2), start=(0, stagger), stop=lo_size)
            recon = upConv(image=tmp_recon, filt=lo_filter, edge_type=edge_type, step=(2, 1), start=(stagger, 0), stop=output_size)

            bands_recon_dict = {
                0: [{'filt': lo_filter.T, 'start': (0, stagger), 'stop': hi_size},
                    {'filt': hi_filter, 'start': (1, 0)}],
                1: [{'filt': hi_filter.T, 'start': (0, 1), 'stop': lo_size},
                    {'filt': lo_filter, 'start': (stagger, 0)}],
                2: [{'filt': hi_filter.T, 'start': (0, 1), 'stop': hi_size},
                    {'filt': hi_filter, 'start': (1, 0)}],
            }

            for band in range(self.num_orientations):
                if (lev, band) in recon_keys:
                    tmp_recon = upConv(image=self.pyr_coeffs[(lev, band)], edge_type=edge_type, step=(1, 2), **bands_recon_dict[band][0])
                    recon += upConv(image=tmp_recon, edge_type=edge_type, step=(2, 1), stop=output_size, **bands_recon_dict[band][1])

        return recon

    def recon_pyr(self, filter_name=None, edge_type=None, levels='all', bands='all'):
        """Reconstruct the input image using pyramid coefficients.

        This function reconstructs the input image using pyramid coefficients.

        Parameters
        ----------
        filter_name : {None, 'binomN', 'haar', 'qmf8', 'qmf12', 'qmf16', 'daub2', 'daub3', 'daub4',
                       'qmf5', 'qmf9', 'qmf13'}
            name of filter to use for reconstruction. All scaled so L-2 norm is 1.0

            * None (default) - use `self.filter_name`, the filter used to construct the pyramid.
            * `'binomN'` - binomial coefficient filter of order N-1
            * `'haar'` - Haar wavelet
            * `'qmf8'`, `'qmf12'`, `'qmf16'` - Symmetric Quadrature Mirror Filters [1]_
            * `'daub2'`, `'daub3'`, `'daub4'` - Daubechies wavelet [2]_
            * `'qmf5'`, `'qmf9'`, `'qmf13'`   - Symmetric Quadrature Mirror Filters [3]_, [4]_
        edge_type : {None, 'circular', 'reflect1', 'reflect2', 'repeat', 'zero', 'extend',
                     'dont-compute'}
            Specifies how to handle edges. Options are:

            * None (default) - use `self.edge_type`, the edge_type used to construct the pyramid
            * `'circular'` - circular convolution
            * `'reflect1'` - reflect about the edge pixels
            * `'reflect2'` - reflect, doubling the edge pixels
            * `'repeat'` - repeat the edge pixels
            * `'zero'` - assume values of zero outside image boundary
            * `'extend'` - reflect and inverts
            * `'dont-compute'` - zero output when filter overhangs imput boundaries.
        levels : `list`, `int`,  or {`'all'`, `'residual_highpass'`}
            If `list` should contain some subset of integers from `0` to `self.num_scales-1`
            (inclusive) and `'residual_lowpass'`. If `'all'`, returned value will contain all
            valid levels. Otherwise, must be one of the valid levels.
        bands : `list`, `int`, or `'all'`.
            If list, should contain some subset of integers from `0` to `self.num_orientations-1`.
            If `'all'`, returned value will contain all valid orientations. Otherwise, must be one
            of the valid orientations.

        Returns
        -------
        recon : `np.array`
            The reconstructed image.
        """
        # Optional args

        if filter_name is None:
            lo_filter = self.filters['lo_filter']
            hi_filter = self.filters['hi_filter']
            stagger = self.stagger
        else:
            lo_filter = parse_filter(filter_name, normalize=False)
            hi_filter = WaveletPyramid._modulate_flip(lo_filter)
            stagger = (lo_filter.size + 1) % 2

        if edge_type is None:
            edges = self.edge_type
        else:
            edges = edge_type

        recon_keys = self._recon_keys(levels, bands)

        # initialize reconstruction
        if 'residual_lowpass' in recon_keys:
            recon = self.pyr_coeffs['residual_lowpass']
        else:
            recon = np.zeros_like(self.pyr_coeffs['residual_lowpass'])

        for lev in reversed(range(self.num_scales)):
            if self.num_orientations == 1:
                if lev == 0:
                    output_size = self.image.shape
                else:
                    output_size = self.pyr_size[(lev-1, 0)]
            else:
                output_size = (self.pyr_size[(lev, 0)][0] + self.pyr_size[(lev, 1)][0],
                               self.pyr_size[(lev, 0)][1] + self.pyr_size[(lev, 1)][1])
            recon = self._recon_prev(recon, lev, recon_keys, output_size, lo_filter,
                                     hi_filter, edges, stagger)

        return recon
